# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.time_trace import current_trace
from newrelic.common.object_wrapper import wrap_function_wrapper
from newrelic.core.context import ContextOf, context_wrapper_async


def _bind_thread_handler(loop, source_task, *args, **kwargs):
    return source_task


def thread_handler_wrapper(wrapped, instance, args, kwargs):
    task = _bind_thread_handler(*args, **kwargs)
    with ContextOf(trace_cache_id=id(task), strict=False):
        return wrapped(*args, **kwargs)


def main_wrap_wrapper(wrapped, instance, args, kwargs):
    awaitable = wrapped(*args, **kwargs)
    return context_wrapper_async(awaitable, current_trace(), strict=False)


def instrument_asgiref_sync(module):
    wrap_function_wrapper(module, "SyncToAsync.thread_handler", thread_handler_wrapper)
    wrap_function_wrapper(module, "AsyncToSync.main_wrap", main_wrap_wrapper)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import newrelic.packages.six as six


if hasattr(inspect, 'iscoroutinefunction'):
    def is_coroutine_function(wrapped):
        return inspect.iscoroutinefunction(wrapped)
else:
    def is_coroutine_function(wrapped):
        return False


if six.PY3:
    def is_asyncio_coroutine(wrapped):
        """Return True if func is a decorated coroutine function."""
        return getattr(wrapped, '_is_coroutine', None) is not None
else:
    def is_asyncio_coroutine(wrapped):
        return False


def is_generator_function(wrapped):
    return inspect.isgeneratorfunction(wrapped)


def _iscoroutinefunction_tornado(fn):
    return hasattr(fn, '__tornado_coroutine__')


def is_coroutine_callable(wrapped):
    return is_coroutine_function(wrapped) or is_coroutine_function(getattr(wrapped, "__call__", None))



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Instrumentation module for CherryPy framework.

"""

# TODO
#
# * We don't track time spent in a user supplied error response callback.
#   We do track time in the handle_error() function which calls it though.
#   Because the error_response attribute of the request object could be
#   updated dynamically, is tricky to handle. We would need to replace
#   error_response attribute at the time the request is created, with a
#   data descriptor which detects when it is set and saves away value, and
#   then wraps the value with a function trace when later access. May
#   have to deal with someone overriding the class attribute version of
#   error_response as well, which means need a class level data descriptor
#   which gets trickier as it isn't bound to an instance and has to do
#   tricks to store the error response handler being set against the
#   instance if set via an instance.
#
# * We don't track time spent in hook functions which may be registered
#   for events such as before_handler, on_end_request etc.
#
# * We don't handle any sub dispatching that may be occurring due to the
#   use of XMLRPCDispatcher.

from newrelic.api.error_trace import wrap_error_trace
from newrelic.api.function_trace import FunctionTrace, wrap_function_trace
from newrelic.api.time_trace import notice_error
from newrelic.api.transaction import current_transaction
from newrelic.api.wsgi_application import wrap_wsgi_application
from newrelic.common.object_names import callable_name
from newrelic.common.object_wrapper import (
    ObjectProxy,
    function_wrapper,
    wrap_function_wrapper,
)


def framework_details():
    import cherrypy

    return ("CherryPy", getattr(cherrypy, "__version__", None))


def status_code(exc, value, tb):
    from cherrypy import HTTPError, HTTPRedirect

    # Ignore certain exceptions based on HTTP status codes.

    if isinstance(value, (HTTPError, HTTPRedirect)):
        # In the case of an HTTPRedirect, value will not have a code attr.
        # In the case of an HTTPError that is malformed (e.g.
        # HTTPError("10 Bad error")), value will not have a code attr.
        # In both of those cases, we fall back to value.status
        code = getattr(value, "code", value.status)
        return code

    return None


@function_wrapper
def handler_wrapper(wrapped, instance, args, kwargs):
    transaction = current_transaction()

    # Name the web transaction after the handler function.

    name = callable_name(wrapped)
    transaction.set_transaction_name(name=name)

    # Track how long is spent in the handler and record any exceptions
    # except those which are used for controlling the actions of the
    # application.

    with FunctionTrace(name=name):
        return wrapped(*args, **kwargs)


class ResourceProxy(ObjectProxy):
    def __getattr__(self, name):
        # Methods on the wrapped object corresponding to the HTTP
        # method will always be upper case. Wrap the method when
        # returned with the handler wrapper.

        attr = super(ResourceProxy, self).__getattr__(name)
        return handler_wrapper(attr) if name.isupper() else attr


def wrapper_Dispatcher_find_handler(wrapped, instance, args, kwargs):
    transaction = current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    try:
        # Call the original wrapped function to find the handler.

        obj, vpath = wrapped(*args, **kwargs)

    except:  # Catch all
        # Can end up here when a custom _cp_dispatch() method is
        # used and that raises an exception.
        notice_error(status_code=status_code)
        raise

    if obj:
        if instance.__class__.__name__ == "MethodDispatcher":
            # We initially name the web transaction as if the
            # corresponding method for the HTTP method will not
            # be found and the request will not be allowed. This
            # will be overridden with the actual handler name
            # when the subsequent wrapper around the handler is
            # executed.

            transaction.set_transaction_name("405", group="StatusCode")

            # We have to use a custom object proxy here in order
            # to intercept accesses made by the dispatcher on the
            # returned object, after this function returns, to
            # retrieve the method of the object corresponding to
            # the HTTP method. For each such access we wrap what
            # is returned in the handler wrapper.

            obj = ResourceProxy(obj)

        else:
            # Should be the actual handler, wrap it with the
            # handler wrapper.

            obj = handler_wrapper(obj)

    else:
        # No handler could be found so name the web transaction
        # after the 404 status code.

        transaction.set_transaction_name("404", group="StatusCode")

    return obj, vpath


def wrapper_RoutesDispatcher_find_handler(wrapped, instance, args, kwargs):
    transaction = current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    try:
        # Call the original wrapped function to find the handler.

        handler = wrapped(*args, **kwargs)

    except:  # Catch all
        # Can end up here when the URL was invalid in some way.
        notice_error(status_code=status_code)
        raise

    if handler:
        # Should be the actual handler, wrap it with the handler
        # wrapper.

        handler = handler_wrapper(handler)

    else:
        # No handler could be found so name the web transaction
        # after the 404 status code.

        transaction.set_transaction_name("404", group="StatusCode")

    return handler


def instrument_cherrypy__cpreqbody(module):
    wrap_function_trace(module, "process_multipart")
    wrap_function_trace(module, "process_multipart_form_data")


def instrument_cherrypy__cprequest(module):
    wrap_function_trace(module, "Request.handle_error")


def instrument_cherrypy__cpdispatch(module):
    wrap_function_wrapper(module, "Dispatcher.find_handler", wrapper_Dispatcher_find_handler)
    wrap_function_wrapper(module, "RoutesDispatcher.find_handler", wrapper_RoutesDispatcher_find_handler)
    wrap_error_trace(
        module, "PageHandler.__call__", ignore=["cherrypy._cperror:InternalRedirect"], status_code=status_code
    )


def instrument_cherrypy__cpwsgi(module):
    wrap_wsgi_application(module, "CPWSGIApp.__call__", framework=framework_details())


def instrument_cherrypy__cptree(module):
    wrap_wsgi_application(module, "Application.__call__", framework=framework_details())



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import logging
import threading

try:
    from newrelic.core.infinite_tracing_pb2 import AttributeValue
except:
    AttributeValue = None

_logger = logging.getLogger(__name__)


class StreamBuffer(object):
    def __init__(self, maxlen):
        self._queue = collections.deque(maxlen=maxlen)
        self._notify = self.condition()
        self._shutdown = False
        self._seen = 0
        self._dropped = 0

    @staticmethod
    def condition(*args, **kwargs):
        return threading.Condition(*args, **kwargs)

    def shutdown(self):
        with self._notify:
            self._shutdown = True
            self._notify.notify_all()

    def put(self, item):
        with self._notify:
            if self._shutdown:
                return

            self._seen += 1

            # NOTE: dropped can be over-counted as the queue approaches
            # capacity while data is still being transmitted.
            #
            # This is because the length of the queue can be changing as it's
            # being measured.
            if len(self._queue) >= self._queue.maxlen:
                self._dropped += 1

            self._queue.append(item)
            self._notify.notify_all()

    def stats(self):
        with self._notify:
            seen, dropped = self._seen, self._dropped
            self._seen, self._dropped = 0, 0

        return seen, dropped

    def __iter__(self):
        return StreamBufferIterator(self)


class StreamBufferIterator(object):
    def __init__(self, stream_buffer):
        self.stream_buffer = stream_buffer
        self._notify = self.stream_buffer._notify
        self._shutdown = False
        self._stream = None

    def shutdown(self):
        with self._notify:
            self._shutdown = True
            self._notify.notify_all()

    def stream_closed(self):
        return self._shutdown or self.stream_buffer._shutdown or (self._stream and self._stream.done())

    def __next__(self):
        with self._notify:
            while True:
                # When a gRPC stream receives a server side disconnect (usually in the form of an OK code)
                # the item it is waiting to consume from the iterator will not be sent, and will inevitably
                # be lost. To prevent this, StopIteration is raised by shutting down the iterator and
                # notifying to allow the thread to exit. Iterators cannot be reused or race conditions may
                # occur between iterator shutdown and restart, so a new iterator must be created from the
                # streaming buffer.
                if self.stream_closed():
                    _logger.debug("gRPC stream is closed. Shutting down and refusing to iterate.")
                    if not self._shutdown:
                        self.shutdown()
                    raise StopIteration

                try:
                    return self.stream_buffer._queue.popleft()
                except IndexError:
                    pass

                if not self.stream_closed() and not self.stream_buffer._queue:
                    self._notify.wait()

    next = __next__

    def __iter__(self):
        return self


class SpanProtoAttrs(dict):
    def __init__(self, *args, **kwargs):
        super(SpanProtoAttrs, self).__init__()
        if args:
            arg = args[0]
            if len(args) > 1:
                raise TypeError("SpanProtoAttrs expected at most 1 argument, got %d", len(args))
            elif hasattr(arg, "keys"):
                for k in arg:
                    self[k] = arg[k]
            else:
                for k, v in arg:
                    self[k] = v

        for k in kwargs:
            self[k] = kwargs[k]

    def __setitem__(self, key, value):
        super(SpanProtoAttrs, self).__setitem__(key, SpanProtoAttrs.get_attribute_value(value))

    def copy(self):
        copy = SpanProtoAttrs()
        copy.update(self)
        return copy

    @staticmethod
    def get_attribute_value(value):
        if isinstance(value, bool):
            return AttributeValue(bool_value=value)
        elif isinstance(value, float):
            return AttributeValue(double_value=value)
        elif isinstance(value, int):
            return AttributeValue(int_value=value)
        else:
            return AttributeValue(string_value=str(value))



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cProfile
import functools
import os
import threading
import time

from newrelic.common.object_wrapper import FunctionWrapper, wrap_object

class FunctionProfile(object):

    def __init__(self, profile):
        self.profile = profile

    def __enter__(self):
        self.profile.enable()
        return self

    def __exit__(self, exc, value, tb):
        self.profile.disable()
        pass

class FunctionProfileSession(object):

    def __init__(self, filename, delay=1.0, checkpoint=30):
        self.filename = filename % { 'pid': os.getpid() }
        self.delay = delay
        self.checkpoint = checkpoint

        self.lock = threading.Lock()
        self.profile = cProfile.Profile()

        self.last = time.time() - delay

        self.active = False
        self.count = 0

    def __call__(self, wrapped, instance, args, kwargs):
        with self.lock:
            if self.active:
                return wrapped(*args, **kwargs)

            if time.time() - self.last < self.delay:
                return wrapped(*args, **kwargs)

            self.active = True
            self.count += 1

        try:
            with FunctionProfile(self.profile):
                result = wrapped(*args, **kwargs)

            if (self.count % self.checkpoint) == 0:
                self.profile.dump_stats(self.filename)

            return result

        finally:
            self.last = time.time()
            self.active = False

def FunctionProfileWrapper(wrapped, filename, delay=1.0, checkpoint=30):
    wrapper = FunctionProfileSession(filename, delay, checkpoint)
    return FunctionWrapper(wrapped, wrapper)

def function_profile(filename, delay=1.0, checkpoint=30):
    return functools.partial(FunctionProfileWrapper, filename=filename,
            delay=delay, checkpoint=checkpoint)

def wrap_function_profile(module, object_path, filename, delay=1.0,
        checkpoint=30):
    wrap_object(module, object_path, FunctionProfileWrapper,
            (filename, delay, checkpoint))



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import newrelic.api.function_trace
import newrelic.api.object_wrapper

class TemplateRenderWrapper(object):

    def __init__(self, wrapped):
        self.__wrapped = wrapped

    def __getattr__(self, name):
        return getattr(self.__wrapped, name)

    def __get__(self, instance, klass):
        if instance is None:
            return self
        descriptor = self.__wrapped.__get__(instance, klass)
        return self.__class__(descriptor)

    def __call__(self, template, *args, **kwargs):
        transaction = newrelic.api.transaction.current_transaction()
        if transaction:
            if hasattr(template, 'filename'):
                name = template.filename or '<template>'
                with newrelic.api.function_trace.FunctionTrace(
                        name=name, group='Template/Render'):
                    return self.__wrapped(template, *args, **kwargs)
            else:
                return self.__wrapped(template, *args, **kwargs)
        else:
            return self.__wrapped(template, *args, **kwargs)

def instrument_mako_runtime(module):

    newrelic.api.object_wrapper.wrap_object(module,
            '_render', TemplateRenderWrapper)

def instrument_mako_template(module):

    def template_filename(template, text, filename, *args):
        return filename

    newrelic.api.function_trace.wrap_function_trace(
            module, '_compile_text',
            name=template_filename, group='Template/Compile')
    newrelic.api.function_trace.wrap_function_trace(
            module, '_compile_module_file',
            name=template_filename, group='Template/Compile')



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.database_trace import register_database_client
from newrelic.common.object_wrapper import wrap_object, ObjectProxy

from newrelic.hooks.database_dbapi2 import ConnectionFactory


class SessionPoolProxy(ObjectProxy):
    def __init__(self, pool, dbapi2_module):
        super(SessionPoolProxy, self).__init__(pool)
        self._nr_dbapi2_module = dbapi2_module

    def acquire(self, *args, **kwargs):
        return ConnectionFactory(
                self.__wrapped__.acquire,
                self._nr_dbapi2_module)(*args, **kwargs)

    def drop(self, connection, *args, **kwargs):
        if isinstance(connection, ObjectProxy):
            connection = connection.__wrapped__

        return self.__wrapped__.drop(connection, *args, **kwargs)

    def release(self, connection, *args, **kwargs):
        if isinstance(connection, ObjectProxy):
            connection = connection.__wrapped__

        return self.__wrapped__.release(connection, *args, **kwargs)


class CreateSessionPoolProxy(ObjectProxy):

    def __init__(self, pool_init, dbapi2_module):
        super(CreateSessionPoolProxy, self).__init__(pool_init)
        self._nr_dbapi2_module = dbapi2_module

    def __call__(self, *args, **kwargs):
        pool = self.__wrapped__(*args, **kwargs)
        return SessionPoolProxy(pool, self._nr_dbapi2_module)


def instrument_cx_oracle(module):
    register_database_client(module, database_product='Oracle',
            quoting_style='single+oracle')

    wrap_object(module, 'connect', ConnectionFactory, (module,))
    wrap_object(module, 'SessionPool', CreateSessionPoolProxy, (module,))



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gc
import os
import platform
import time
from collections import Counter

from newrelic.common.object_names import callable_name
from newrelic.core.config import global_settings
from newrelic.core.stats_engine import CustomMetrics
from newrelic.samplers.decorators import data_source_factory


@data_source_factory(name="Garbage Collector Metrics")
class _GCDataSource(object):
    def __init__(self, settings, environ):
        self.gc_time_metrics = CustomMetrics()
        self.start_time = 0.0
        self.previous_stats = {}
        self.pid = os.getpid()

    @property
    def enabled(self):
        settings = global_settings()
        if platform.python_implementation() == "PyPy" or not settings:
            return False
        else:
            return settings.gc_runtime_metrics.enabled

    @property
    def top_object_count_limit(self):
        settings = global_settings()
        return settings.gc_runtime_metrics.top_object_count_limit

    def record_gc(self, phase, info):
        if not self.enabled:
            return

        current_generation = info["generation"]

        if phase == "start":
            self.start_time = time.time()
        elif phase == "stop":
            total_time = time.time() - self.start_time
            self.gc_time_metrics.record_custom_metric("GC/time/%d/all" % self.pid, total_time)
            for gen in range(0, 3):
                if gen <= current_generation:
                    self.gc_time_metrics.record_custom_metric("GC/time/%d/%d" % (self.pid, gen), total_time)
                else:
                    self.gc_time_metrics.record_custom_metric("GC/time/%d/%d" % (self.pid, gen), 0)

    def start(self):
        if hasattr(gc, "callbacks"):
            gc.callbacks.append(self.record_gc)

    def stop(self):
        # The callback must be removed before resetting the metrics tables.
        # If it isn't, it's possible to be interrupted by the gc and to have more
        # metrics appear in the table that should be empty.
        if hasattr(gc, "callbacks") and self.record_gc in gc.callbacks:
            gc.callbacks.remove(self.record_gc)

        self.gc_time_metrics.reset_metric_stats()
        self.start_time = 0.0

    def __call__(self):
        if not self.enabled:
            return

        # Record object count in total and per generation
        if hasattr(gc, "get_count"):
            counts = gc.get_count()
            yield ("GC/objects/%d/all" % self.pid, {"count": sum(counts)})
            for gen, count in enumerate(counts):
                yield (
                    "GC/objects/%d/generation/%d" % (self.pid, gen),
                    {"count": count},
                )

        # Record object count for top five types with highest count
        if hasattr(gc, "get_objects"):
            object_types = map(type, gc.get_objects())
            if self.top_object_count_limit > 0:
                highest_types = Counter(object_types).most_common(self.top_object_count_limit)
                for obj_type, count in highest_types:
                    yield (
                        "GC/objects/%d/type/%s" % (self.pid, callable_name(obj_type)),
                        {"count": count},
                    )

        if hasattr(gc, "get_stats"):
            stats_by_gen = gc.get_stats()
            if isinstance(stats_by_gen, list):
                for stat_name in stats_by_gen[0].keys():
                    # Aggregate metrics for /all
                    count = sum(stats[stat_name] for stats in stats_by_gen)
                    previous_value = self.previous_stats.get((stat_name, "all"), 0)
                    self.previous_stats[(stat_name, "all")] = count
                    change_in_value = count - previous_value
                    yield (
                        "GC/%s/%d/all" % (stat_name, self.pid),
                        {"count": change_in_value},
                    )

                    # Breakdowns by generation
                    for gen, stats in enumerate(stats_by_gen):
                        previous_value = self.previous_stats.get((stat_name, gen), 0)
                        self.previous_stats[(stat_name, gen)] = stats[stat_name]
                        change_in_value = stats[stat_name] - previous_value

                        yield (
                            "GC/%s/%d/%d" % (stat_name, self.pid, gen),
                            {"count": change_in_value},
                        )

        # In order to avoid a concurrency issue with getting interrupted by the
        # garbage collector, we save a reference to the old metrics table, and overwrite
        # self.gc_time_metrics with a new empty table via reset_metric_stats().
        # This guards against losing data points, or having inconsistent data points
        # reported between /all and the totals of /generation/%d metrics.
        gc_time_metrics = self.gc_time_metrics.metrics()
        self.gc_time_metrics.reset_metric_stats()

        for metric in gc_time_metrics:
            raw_metric = metric[1]
            yield metric[0], {
                "count": raw_metric.call_count,
                "total": raw_metric.total_call_time,
                "min": raw_metric.min_call_time,
                "max": raw_metric.max_call_time,
                "sum_of_squares": raw_metric.sum_of_squares,
            }


garbage_collector_data_source = _GCDataSource



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module implements utilities for context propagation for tracing across threads.
"""

import logging

from newrelic.common.object_wrapper import function_wrapper
from newrelic.core.trace_cache import trace_cache

_logger = logging.getLogger(__name__)


class ContextOf(object):
    def __init__(self, trace=None, request=None, trace_cache_id=None, strict=True):
        self.trace = None
        self.trace_cache = trace_cache()
        self.thread_id = None
        self.restore = None
        self.should_restore = False

        def log_propagation_failure(s):
            if strict:
                _logger.error(
                    "Runtime instrumentation error. Request context propagation failed. %s Report this issue to New Relic support.",
                    s,
                )
            else:
                _logger.debug(
                    "Request context propagation failed. %s This may be an issue if there's an active transaction. Consult with New Relic support if further issues arise.",
                    s,
                )

        # Extract trace if possible, else leave as None for safety
        if trace is None and request is None and trace_cache_id is None:
            if strict:
                log_propagation_failure("No trace or request provided.")
        elif trace is not None:
            self.trace = trace
        elif trace_cache_id is not None:
            self.trace = self.trace_cache._cache.get(trace_cache_id, None)
            if self.trace is None:
                log_propagation_failure("No trace with id %d." % trace_cache_id)
        elif hasattr(request, "_nr_trace") and request._nr_trace is not None:
            # Unpack traces from objects patched with them
            self.trace = request._nr_trace
        else:
            log_propagation_failure("No context attached to request.")

    def __enter__(self):
        if self.trace:
            self.thread_id = self.trace_cache.current_thread_id()

            # Save previous cache contents
            self.restore = self.trace_cache._cache.get(self.thread_id, None)
            self.should_restore = True

            # Set context in trace cache
            self.trace_cache._cache[self.thread_id] = self.trace

        return self

    def __exit__(self, exc, value, tb):
        if self.should_restore:
            if self.restore is not None:
                # Restore previous contents
                self.trace_cache._cache[self.thread_id] = self.restore
            else:
                # Remove entry from cache
                self.trace_cache._cache.pop(self.thread_id)


def context_wrapper(func, trace=None, request=None, trace_cache_id=None, strict=True):
    @function_wrapper
    def _context_wrapper(wrapped, instance, args, kwargs):
        with ContextOf(trace=trace, request=request, trace_cache_id=trace_cache_id, strict=strict):
            return wrapped(*args, **kwargs)

    return _context_wrapper(func)


async def context_wrapper_async(awaitable, trace=None, request=None, trace_cache_id=None, strict=True):
    with ContextOf(trace=trace, request=request, trace_cache_id=trace_cache_id, strict=strict):
        return await awaitable


def current_thread_id():
    return trace_cache().current_thread_id()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import logging
import os
import threading
import time
import zlib
from collections import defaultdict, deque

import newrelic
import newrelic.packages.six as six
from newrelic.common.encoding_utils import json_encode
from newrelic.core.config import global_settings
from newrelic.core.trace_cache import trace_cache

try:
    from sys import intern
except ImportError:
    pass

_logger = logging.getLogger(__name__)

AGENT_PACKAGE_DIRECTORY = os.path.dirname(newrelic.__file__) + "/"


class SessionState(object):
    RUNNING = 1
    FINISHED = 2


def format_stack_trace(frame, thread_category):
    """Formats the frame obj into a list of stack trace tuples."""

    stack_trace = deque()

    while frame:
        # The value frame.f_code.co_firstlineno is the first line of
        # code in the file for the specified function. The value
        # frame.f_lineno is the actual line which is being executed
        # at the time the stack frame was being viewed.

        code = frame.f_code

        filename = intern(code.co_filename)
        func_name = intern(code.co_name)
        first_line = code.co_firstlineno

        real_line = frame.f_lineno

        # Set ourselves up to process next frame back up the stack.

        frame = frame.f_back

        # So as to make it more obvious to the user as to what their
        # code is doing, we drop out stack frames related to the
        # agent instrumentation. Don't do this for the agent threads
        # though as we still need to seem them in that case so can
        # debug what the agent itself is doing.

        if thread_category != "AGENT" and filename.startswith(AGENT_PACKAGE_DIRECTORY):
            continue

        if not stack_trace:
            # Add the fake leaf node with line number of where the
            # code was executing at the point of the sample. This
            # could be actual Python code within the function, or
            # more likely showing the point where a call is being
            # made into a C function wrapped as Python object. The
            # latter can occur because we will not see stack frames
            # when calling into C functions.

            stack_trace.appendleft((filename, func_name, real_line, real_line))

        # Add the actual node for the function being called at this
        # level in the stack frames.

        stack_trace.appendleft((filename, func_name, first_line, real_line))

    return stack_trace


def collect_stack_traces(include_nr_threads=False):
    """Generator that yields the (thread category, stack trace) of all the
    python threads.

    """
    for (txn, thread_id, thread_category, frame) in trace_cache().active_threads():

        # Skip NR Threads unless explicitly requested.

        if (thread_category == "AGENT") and (not include_nr_threads):
            continue

        stack_trace = format_stack_trace(frame, thread_category)

        # Skip over empty stack traces. This is merely for optimization.
        #
        # It saves us from adding an empty deque to the txn obj, which will be
        # discarded later on during call tree merge.

        if not stack_trace:
            continue

        yield thread_category, stack_trace


class ProfileSessionManager(object):
    """Singleton class that manages multiple profile sessions. Do NOT
    instantiate directly from this class. Instead use profile_session_manager()

    """

    _lock = threading.Lock()
    _instance = None

    @staticmethod
    def singleton():
        with ProfileSessionManager._lock:
            if ProfileSessionManager._instance is None:
                ProfileSessionManager._instance = ProfileSessionManager()

        return ProfileSessionManager._instance

    def __init__(self):
        self.full_profile_session = None

        # Name of the application that requested the full_profile session.

        self.full_profile_app = None

        self.finished_sessions = defaultdict(list)

        self._profiler_shutdown = threading.Event()
        self._profiler_thread = None
        self._profiler_thread_running = False
        self._lock = threading.Lock()
        self.profile_agent_code = False
        self.sample_period_s = 0.1

    def start_profile_session(self, app_name, profile_id, stop_time, sample_period_s=0.1, profile_agent_code=False):
        """Start a new profiler session. If a full_profiler is already
        running, do nothing and return false.

        """

        # XXX Only one instance of this at the moment. Also only one harvest
        # thread so do not need thread locking. If we move to multiple harvest
        # threads, would need locking.

        # Only one full profile session can run at any given time.

        if self.full_profile_session:
            # log an error message
            return False

        # Acquire thread lock before updating the data structures. This method
        # is invoked from the harvest thread and this ensures the variables are
        # not being updated concurrently by the profiler thread.

        with self._lock:

            self.profile_agent_code = profile_agent_code
            self.sample_period_s = sample_period_s
            self.full_profile_session = ProfileSession(profile_id, stop_time)
            self.full_profile_app = app_name

            # Create a background thread to collect stack traces. Do this only
            # if a background thread doesn't already exist.

            if not self._profiler_thread_running:
                self._profiler_thread = threading.Thread(target=self._profiler_loop, name="NR-Profiler-Thread")
                self._profiler_thread.daemon = True

                self._profiler_thread.start()
                self._profiler_thread_running = True

        return True

    def stop_profile_session(self, app_name):
        """Stop a profiler session and return True when successful. Set key_txn
        to None to stop the full_profile_session. Returns False if no profiler
        sessions were stopped.

        """

        # Grab the lock before updating the profile sessions. This is to
        # make sure that we're not updating the data structures while the
        # harvest thread is starting/stopping new sessions.

        with self._lock:
            if (self.full_profile_session is not None) and (app_name == self.full_profile_app):
                self.full_profile_session.state = SessionState.FINISHED
                self.full_profile_session.actual_stop_time_s = time.time()
                self.finished_sessions[app_name].append(self.full_profile_session)
                self.full_profile_session = None
                self.full_profile_app = None

            else:
                # log an error message
                return False

    def profile_data(self, app_name):
        """Generator that yields the profile data in the form of call tree.
        Data from full profiler will only be returned when the profiler has
        finished. When no data is available None will be returned.

        """
        with self._lock:
            for session in self.finished_sessions[app_name]:
                _logger.debug(
                    "Reporting final thread profiling data for "
                    "%d transactions over a period of %.2f seconds "
                    "and %d samples.",
                    session.transaction_count,
                    time.time() - session.start_time_s,
                    session.sample_count,
                )

                yield session.profile_data()

            # After reporting data on the finished_sessions empty the
            # finished_sessions list.

            self.finished_sessions.pop(app_name)

    def _profiler_loop(self):
        """Infinite loop that wakes up periodically to collect stack traces,
        merge it into call tree if necessary, finally update the state of all
        the active profile sessions.

        """

        while True:

            for category, stack in collect_stack_traces(self.profile_agent_code):

                # Merge the stack_trace to the call tree only for
                # full_profile_session.

                if self.full_profile_session:
                    self.full_profile_session.update_call_tree(category, stack)

            self.update_profile_sessions()

            # Stop the profiler thread if there are no profile sessions.

            if self.full_profile_session is None:
                self._profiler_thread_running = False
                return

            self._profiler_shutdown.wait(self.sample_period_s)

    def update_profile_sessions(self):
        """Check the current time and decide if any of the profile sessions
        have expired and move it to the finished_sessions list.

        """

        if self.full_profile_session:
            self.full_profile_session.sample_count += 1
            if time.time() >= self.full_profile_session.stop_time_s:
                self.stop_profile_session(self.full_profile_app)
                _logger.info("Finished thread profiling session.")

    def shutdown(self, app_name):
        """Stop all profile sessions running on the given app_name."""

        # Check if we need to stop the full profiler.

        if app_name == self.full_profile_app:
            self.stop_profile_session(app_name)

        return True


class ProfileSession(object):
    def __init__(self, profile_id, stop_time):
        self.profile_id = profile_id
        self.start_time_s = time.time()
        self.stop_time_s = stop_time
        self.actual_stop_time_s = 0
        self.state = SessionState.RUNNING
        self.reset_profile_data()

    def reset_profile_data(self):
        self.call_buckets = {"REQUEST": {}, "AGENT": {}, "BACKGROUND": {}, "OTHER": {}}
        self._node_list = []
        self.start_time_s = time.time()
        self.sample_count = 0
        self.transaction_count = 0

    def update_call_tree(self, bucket_type, stack_trace):
        """Merge a single call stack trace into a call tree bucket. If
        no appropriate call tree is found then create a new call tree.
        An appropriate call tree will have the same root node as the
        last method in the stack trace.

        """

        self.transaction_count += 1

        depth = 1
        try:
            bucket = self.call_buckets[bucket_type]
        except KeyError:
            return False

        for method in stack_trace:
            call_tree = bucket.get(method)

            if call_tree is None:
                call_tree = CallTree(method, depth=depth)
                self._node_list.append(call_tree)
                bucket[method] = call_tree

            call_tree.call_count += 1

            # The call depth is incremented on each recursive call so we
            # know the depth of the call stack. We use this later when
            # pruning nodes if go over the limit. Specifically, the deepest
            # and least used nodes will be prune first.

            bucket = call_tree.children
            depth += 1

        return True

    def _prune_call_trees(self, limit):
        """Prune the number of profile nodes we send up to the data
        collector down to the specified limit. Done to ensure not
        sending so much data that gets reject for being over size limit.

        """

        if len(self._node_list) <= limit:
            return

        # We sort the profile nodes based on call count, but also take
        # into consideration the depth of the node in the call tree.
        # Based on sort order, we then ignore any nodes over our limit.
        #
        # We include depth as that way we try and trim the deepest and
        # least visited leaf nodes first. If we don't do this, then
        # depending on how sorting orders nodes with same call count, we
        # could ignore a parent node high up in call chain even though
        # children weren't being ignored and so effectively ignore more
        # than the minimum we need to. Granted this would only occur
        # where was a linear call tree where all had the same call count,
        # such as may occur with recursion.
        #
        # Also note that we still can actually end up with less nodes in
        # the end being displayed in the UI than the limit being applied
        # even though we initially cutoff at the limit. This is because
        # we are looking at nodes from different categories before they
        # have been merged together. If a node appears at same relative
        # position in multiple categories, then when displaying multiple
        # categories in UI, the duplicates only appear as one after the
        # UI merges them.

        self._node_list.sort(key=lambda x: (x.call_count, -x.depth), reverse=True)

        for node in self._node_list[limit:]:
            node.ignore = True

    def profile_data(self):

        # Generic profiling sessions have to wait for completion before
        # reporting data.

        if self.state == SessionState.RUNNING:
            return None

        # We prune the number of nodes sent if we are over the specified
        # limit. This is just to avoid having the response be too large
        # and get rejected by the data collector.

        settings = global_settings()
        self._prune_call_trees(settings.agent_limits.thread_profiler_nodes)

        flat_tree = {}
        thread_count = 0

        for category, bucket in six.iteritems(self.call_buckets):

            # Only flatten buckets that have data in them. No need to send
            # empty buckets.

            if bucket:
                flat_tree[category] = [x.flatten() for x in bucket.values()]
                thread_count += len(bucket)

        # Construct the actual final data for sending. The actual call
        # data is turned into JSON, compressed and then base64 encoded at
        # this point to cut its size.

        if settings.debug.log_thread_profile_payload:
            _logger.debug("Encoding thread profile data where payload=%r.", flat_tree)

        json_call_tree = json_encode(flat_tree)

        level = settings.agent_limits.data_compression_level
        level = level or zlib.Z_DEFAULT_COMPRESSION

        encoded_tree = base64.standard_b64encode(zlib.compress(six.b(json_call_tree), level))

        if six.PY3:
            encoded_tree = encoded_tree.decode("Latin-1")

        profile = [
            [
                self.profile_id,
                self.start_time_s * 1000,
                (self.actual_stop_time_s or time.time()) * 1000,
                self.sample_count,
                encoded_tree,
                thread_count,
                0,
                None,
            ]
        ]

        # Reset the data structures to default.

        self.reset_profile_data()
        return profile


class CallTree(object):
    def __init__(self, method_data, call_count=0, depth=1):
        self.method_data = method_data
        self.call_count = call_count
        self.children = {}

        self.depth = depth
        self.ignore = False

    def flatten(self):
        filename, func_name, func_line, exec_line = self.method_data

        # func_line is the first line of a function and exec_line is the line
        # inside that function that is currently being executed.  On the leaf
        # nodes the exec_line will be different from the func_line. Such nodes
        # are labeled with an @ sign in the second element of the tuple.

        if func_line == exec_line:
            method_data = (filename, "@%s#%s" % (func_name, func_line), exec_line)
        else:
            method_data = (filename, "%s#%s" % (func_name, func_line), exec_line)

        return [method_data, self.call_count, 0, [x.flatten() for x in self.children.values() if not x.ignore]]


def profile_session_manager():
    return ProfileSessionManager.singleton()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple

import newrelic.core.trace_node
from newrelic.core.node_mixin import GenericNodeMixin
from newrelic.core.attribute import resolve_user_attributes

from newrelic.packages import six

_RootNode = namedtuple('_RootNode',
        ['name', 'children', 'start_time', 'end_time', 'exclusive',
        'duration', 'guid', 'agent_attributes', 'user_attributes',
        'path', 'trusted_parent_span', 'tracing_vendors',])


class RootNode(_RootNode, GenericNodeMixin):
    def span_event(self, *args, **kwargs):
        span = super(RootNode, self).span_event(*args, **kwargs)
        i_attrs = span[0]
        i_attrs['transaction.name'] = self.path
        i_attrs['nr.entryPoint'] = True
        if self.trusted_parent_span:
            i_attrs['trustedParentId'] = self.trusted_parent_span
        if self.tracing_vendors:
            i_attrs['tracingVendors'] = self.tracing_vendors
        return span

    def trace_node(self, stats, root, connections):

        name = self.path

        start_time = newrelic.core.trace_node.node_start_time(root, self)
        end_time = newrelic.core.trace_node.node_end_time(root, self)

        root.trace_node_count += 1

        children = []

        for child in self.children:
            if root.trace_node_count > root.trace_node_limit:
                break
            children.append(child.trace_node(stats, root, connections))

        params = self.get_trace_segment_params(root.settings)

        return newrelic.core.trace_node.TraceNode(
                start_time=start_time,
                end_time=end_time,
                name=name,
                params=params,
                children=children,
                label=None)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import logging

from newrelic.api.time_trace import TimeTrace, current_trace
from newrelic.common.async_wrapper import async_wrapper
from newrelic.common.object_wrapper import FunctionWrapper, wrap_object
from newrelic.core.database_node import DatabaseNode
from newrelic.core.stack_trace import current_stack

_logger = logging.getLogger(__name__)


def register_database_client(
    dbapi2_module, database_product, quoting_style="single", explain_query=None, explain_stmts=[], instance_info=None
):

    _logger.debug(
        "Registering database client module %r where database "
        "is %r, quoting style is %r, explain query statement is %r and "
        "the SQL statements on which explain plans can be run are %r.",
        dbapi2_module,
        database_product,
        quoting_style,
        explain_query,
        explain_stmts,
    )

    dbapi2_module._nr_database_product = database_product
    dbapi2_module._nr_quoting_style = quoting_style
    dbapi2_module._nr_explain_query = explain_query
    dbapi2_module._nr_explain_stmts = explain_stmts
    dbapi2_module._nr_instance_info = instance_info
    dbapi2_module._nr_datastore_instance_feature_flag = False


def enable_datastore_instance_feature(dbapi2_module):
    dbapi2_module._nr_datastore_instance_feature_flag = True


class DatabaseTrace(TimeTrace):

    __async_explain_plan_logged = False

    def __init__(
        self,
        sql,
        dbapi2_module=None,
        connect_params=None,
        cursor_params=None,
        sql_parameters=None,
        execute_params=None,
        host=None,
        port_path_or_id=None,
        database_name=None,
        **kwargs
    ):
        parent = None
        if kwargs:
            if len(kwargs) > 1:
                raise TypeError("Invalid keyword arguments:", kwargs)
            parent = kwargs["parent"]
        super(DatabaseTrace, self).__init__(parent)

        self.sql = sql

        self.dbapi2_module = dbapi2_module

        self.connect_params = connect_params
        self.cursor_params = cursor_params
        self.sql_parameters = sql_parameters
        self.execute_params = execute_params
        self.host = host
        self.port_path_or_id = port_path_or_id
        self.database_name = database_name

    def __enter__(self):
        result = super(DatabaseTrace, self).__enter__()
        if result and self.transaction:
            self.sql = self.transaction._intern_string(self.sql)
        return result

    def __repr__(self):
        return "<%s object at 0x%x %s>" % (
            self.__class__.__name__,
            id(self),
            dict(sql=self.sql, dbapi2_module=self.dbapi2_module),
        )

    @property
    def is_async_mode(self):
        # Check for `async=1` keyword argument in connect_params, which
        # indicates that psycopg2 driver is being used in async mode.

        # In python 3.7+ the arg will be 'async_' because 'async' is a keyword
        # However, note that psycopg2 v2.7+ allows either one as aliases, so
        # we have to check both.

        try:
            _, kwargs = self.connect_params
        except TypeError:
            return False
        else:
            return ("async" in kwargs and kwargs["async"]) or ("async_" in kwargs and kwargs["async_"])

    def _log_async_warning(self):
        # Only log the warning the first time.

        if not DatabaseTrace.__async_explain_plan_logged:
            DatabaseTrace.__async_explain_plan_logged = True
            _logger.warning(
                "Explain plans are not supported for queries made over database connections in asynchronous mode."
            )

    def finalize_data(self, transaction, exc=None, value=None, tb=None):
        self.stack_trace = None

        connect_params = None
        cursor_params = None
        sql_parameters = None
        execute_params = None
        host = None
        port_path_or_id = None
        database_name = None

        settings = transaction.settings

        if not settings:
            return

        tt = settings.transaction_tracer
        agent_limits = settings.agent_limits
        ds_tracer = settings.datastore_tracer

        # Check settings, so that we only call instance_info when needed.

        instance_enabled = ds_tracer.instance_reporting.enabled
        db_name_enabled = ds_tracer.database_name_reporting.enabled

        if instance_enabled or db_name_enabled:

            if (
                self.dbapi2_module
                and self.connect_params
                and self.dbapi2_module._nr_datastore_instance_feature_flag
                and self.dbapi2_module._nr_instance_info is not None
            ):

                instance_info = self.dbapi2_module._nr_instance_info(*self.connect_params)

                if instance_enabled:
                    host, port_path_or_id, _ = instance_info

                if db_name_enabled:
                    _, _, database_name = instance_info

            else:
                if instance_enabled:
                    host = self.host
                    port_path_or_id = self.port_path_or_id

                if db_name_enabled:
                    database_name = self.database_name

        if tt.enabled and settings.collect_traces and tt.record_sql != "off":
            if self.duration >= tt.stack_trace_threshold:
                if transaction._stack_trace_count < agent_limits.slow_sql_stack_trace:
                    self.stack_trace = [transaction._intern_string(x) for x in current_stack(skip=2)]
                    transaction._stack_trace_count += 1

            if self.is_async_mode and tt.explain_enabled:
                self._log_async_warning()
            else:
                # Only remember all the params for the calls if know
                # there is a chance we will need to do an explain
                # plan. We never allow an explain plan to be done if
                # an exception occurred in doing the query in case
                # doing the explain plan with the same inputs could
                # cause further problems.

                if (
                    exc is None
                    and not self.is_async_mode
                    and tt.explain_enabled
                    and self.duration >= tt.explain_threshold
                    and self.connect_params is not None
                ):
                    if transaction._explain_plan_count < agent_limits.sql_explain_plans:
                        connect_params = self.connect_params
                        cursor_params = self.cursor_params
                        sql_parameters = self.sql_parameters
                        execute_params = self.execute_params
                        transaction._explain_plan_count += 1

        self.sql_format = tt.record_sql

        self.connect_params = connect_params
        self.cursor_params = cursor_params
        self.sql_parameters = sql_parameters
        self.execute_params = execute_params
        self.host = host
        self.port_path_or_id = port_path_or_id
        self.database_name = database_name

    def terminal_node(self):
        return True

    def create_node(self):
        return DatabaseNode(
            dbapi2_module=self.dbapi2_module,
            sql=self.sql,
            children=self.children,
            start_time=self.start_time,
            end_time=self.end_time,
            duration=self.duration,
            exclusive=self.exclusive,
            stack_trace=self.stack_trace,
            sql_format=self.sql_format,
            connect_params=self.connect_params,
            cursor_params=self.cursor_params,
            sql_parameters=self.sql_parameters,
            execute_params=self.execute_params,
            host=self.host,
            port_path_or_id=self.port_path_or_id,
            database_name=self.database_name,
            guid=self.guid,
            agent_attributes=self.agent_attributes,
            user_attributes=self.user_attributes,
        )


def DatabaseTraceWrapper(wrapped, sql, dbapi2_module=None):
    def _nr_database_trace_wrapper_(wrapped, instance, args, kwargs):
        wrapper = async_wrapper(wrapped)
        if not wrapper:
            parent = current_trace()
            if not parent:
                return wrapped(*args, **kwargs)
        else:
            parent = None

        if callable(sql):
            if instance is not None:
                _sql = sql(instance, *args, **kwargs)
            else:
                _sql = sql(*args, **kwargs)
        else:
            _sql = sql

        trace = DatabaseTrace(_sql, dbapi2_module, parent=parent)

        if wrapper:  # pylint: disable=W0125,W0126
            return wrapper(wrapped, trace)(*args, **kwargs)

        with trace:
            return wrapped(*args, **kwargs)

    return FunctionWrapper(wrapped, _nr_database_trace_wrapper_)


def database_trace(sql, dbapi2_module=None):
    return functools.partial(DatabaseTraceWrapper, sql=sql, dbapi2_module=dbapi2_module)


def wrap_database_trace(module, object_path, sql, dbapi2_module=None):
    wrap_object(module, object_path, DatabaseTraceWrapper, (sql, dbapi2_module))



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.common.object_wrapper import wrap_function_wrapper


def framework_details():
    import graphene

    return ("Graphene", getattr(graphene, "__version__", None))


def wrap_schema_init(wrapped, instance, args, kwargs):
    result = wrapped(*args, **kwargs)
    if hasattr(instance, "graphql_schema"):
        instance.graphql_schema._nr_framework = framework_details()
    else:
        instance._nr_framework = framework_details()

    return result  # Should never actually be defined


def instrument_graphene_types_schema(module):
    if hasattr(module, "Schema"):
        wrap_function_wrapper(module, "Schema.__init__", wrap_schema_init)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

from newrelic.admin import command, usage


def _run_validation_test():
    import time

    from newrelic.api.application import application_instance as application
    from newrelic.api.background_task import background_task
    from newrelic.api.error_trace import error_trace
    from newrelic.api.external_trace import external_trace
    from newrelic.api.function_trace import function_trace
    from newrelic.api.transaction import add_custom_parameter
    from newrelic.api.time_trace import notice_error
    from newrelic.api.wsgi_application import wsgi_application

    @external_trace(library='test',
            url='http://localhost/test', method='GET')
    def _external1():
        time.sleep(0.1)

    @function_trace(label='label',
            params={'fun-key-1': '1', 'fun-key-2': 2, 'fun-key-3': 3.0})
    def _function1():
        _external1()

    @function_trace()
    def _function2():
        for i in range(10):
            _function1()

    @error_trace()
    @function_trace()
    def _function3():
        add_custom_parameter('txn-key-1', 1)

        _function4()

        raise RuntimeError('This is a test error and can be ignored.')

    @function_trace()
    def _function4(params=None, application=None):
        try:
            _function5()
        except:
            notice_error(attributes=(params or {
                    'err-key-2': 2, 'err-key-3': 3.0}),
                    application=application)

    @function_trace()
    def _function5():
        raise NotImplementedError(
                'This is a test error and can be ignored.')

    @wsgi_application()
    def _wsgi_application(environ, start_response):
        status = '200 OK'
        output = 'Hello World!'

        response_headers = [('Content-type', 'text/plain'),
                            ('Content-Length', str(len(output)))]
        start_response(status, response_headers)

        for i in range(10):
            _function1()

        _function2()

        time.sleep(1.0)

        try:
            _function3()
        except Exception:
            pass

        return [output]

    @background_task()
    def _background_task():
        for i in range(10):
            _function1()

        _function2()

        time.sleep(0.5)

        try:
            _function3()
        except Exception:
            pass

    def _start_response(*args):
        pass

    _environ = { 'SCRIPT_NAME': '', 'PATH_INFO': '/test',
                 'QUERY_STRING': 'key=value' }

    _iterable = _wsgi_application(_environ, _start_response)
    _iterable.close()

    _background_task()

    _function4(params={'err-key-4': 4, 'err-key-5': 5.0},
            application=application())

_user_message = """
Running Python agent test.

Any significant errors in performing the test will be shown below. If no
errors occurred in the execution of this script and data is still not
reporting through to the UI against the application:

  %(app_name)s

after 5 minutes then check the log file:

  %(log_file)s

for debugging information. Supply the log file to New Relic support if
requesting help with resolving any issues with the test not reporting
data to the New Relic UI.
"""

@command('validate-config', 'config_file [log_file]',
"""Validates the syntax of <config_file>. Also tests connectivity to New
Relic core application by connecting to the account corresponding to the
license key listed in the configuration file, and reporting test data under
the application name 'Python Agent Test'.""")
def validate_config(args):
    import os
    import sys
    import logging
    import time

    if len(args) == 0:
        usage('validate-config')
        sys.exit(1)

    from newrelic.api.application import register_application
    from newrelic.config import initialize
    from newrelic.core.config import global_settings

    if len(args) >= 2:
        log_file = args[1]
    else:
        log_file = '/tmp/python-agent-test.log'

    log_level = logging.DEBUG

    try:
        os.unlink(log_file)
    except Exception:
        pass

    config_file = args[0]
    environment = os.environ.get('NEW_RELIC_ENVIRONMENT')

    if config_file == '-':
        config_file = os.environ.get('NEW_RELIC_CONFIG_FILE')

    initialize(config_file, environment, ignore_errors=False,
            log_file=log_file, log_level=log_level)

    _logger = logging.getLogger(__name__)

    _logger.debug('Starting agent validation.')

    _settings = global_settings()

    app_name = os.environ.get('NEW_RELIC_TEST_APP_NAME', 'Python Agent Test')

    _settings.app_name = app_name
    _settings.transaction_tracer.transaction_threshold = 0
    _settings.shutdown_timeout = 30.0

    _settings.debug.log_malformed_json_data = True
    _settings.debug.log_data_collector_payloads = True
    _settings.debug.log_transaction_trace_payload = True

    print(_user_message % dict(app_name=app_name, log_file=log_file))

    _logger.debug('Register test application.')

    _logger.debug('Collector host is %r.', _settings.host)
    _logger.debug('Collector port is %r.', _settings.port)

    _logger.debug('Proxy scheme is %r.', _settings.proxy_scheme)
    _logger.debug('Proxy host is %r.', _settings.proxy_host)
    _logger.debug('Proxy port is %r.', _settings.proxy_port)
    _logger.debug('Proxy user is %r.', _settings.proxy_user)

    _logger.debug('License key is %r.', _settings.license_key)

    _timeout = 30.0

    _start = time.time()
    _application = register_application(timeout=_timeout)
    _end = time.time()

    _duration = _end - _start

    if not _application.active:
        _logger.error('Unable to register application for test, '
            'connection could not be established within %s seconds.',
            _timeout)
        return

    if hasattr(_application.settings, 'messages'):
        for message in _application.settings.messages:
            if message['message'].startswith('Reporting to:'):
                parts = message['message'].split('Reporting to:')
                url = parts[1].strip()
                print('Registration successful. Reporting to:')
                print()
                print('  %s' % url)
                print()
                break

    _logger.debug('Registration took %s seconds.', _duration)

    _logger.debug('Run the validation test.')

    _run_validation_test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.database_trace import (
    DatabaseTrace,
    enable_datastore_instance_feature,
    register_database_client,
)
from newrelic.api.datastore_trace import DatastoreTrace
from newrelic.common.object_wrapper import ObjectProxy, wrap_function_wrapper


class PostgresApi(object):
    @staticmethod
    def _instance_info(addr, connected_fut, con_params, *args, **kwargs):
        if isinstance(addr, str):
            host = "localhost"
            port = addr
        else:
            host, port = addr

        return (host, str(port), getattr(con_params, "database", None))

    @classmethod
    def instance_info(cls, args, kwargs):
        return cls._instance_info(*args, **kwargs)


register_database_client(
    PostgresApi,
    "Postgres",
    quoting_style="single+dollar",
    instance_info=PostgresApi.instance_info,
)
enable_datastore_instance_feature(PostgresApi)


class ProtocolProxy(ObjectProxy):
    async def bind_execute(self, state, *args, **kwargs):
        with DatabaseTrace(
            state.query,
            dbapi2_module=PostgresApi,
            connect_params=getattr(self, "_nr_connect_params", None),
        ):
            return await self.__wrapped__.bind_execute(state, *args, **kwargs)

    async def bind_execute_many(self, state, *args, **kwargs):
        with DatabaseTrace(
            state.query,
            dbapi2_module=PostgresApi,
            connect_params=getattr(self, "_nr_connect_params", None),
        ):
            return await self.__wrapped__.bind_execute_many(state, *args, **kwargs)

    async def bind(self, state, *args, **kwargs):
        with DatabaseTrace(
            state.query,
            dbapi2_module=PostgresApi,
            connect_params=getattr(self, "_nr_connect_params", None),
        ):
            return await self.__wrapped__.bind(state, *args, **kwargs)

    async def execute(self, state, *args, **kwargs):
        with DatabaseTrace(
            state.query,
            dbapi2_module=PostgresApi,
            connect_params=getattr(self, "_nr_connect_params", None),
        ):
            return await self.__wrapped__.execute(state, *args, **kwargs)

    async def query(self, query, *args, **kwargs):
        with DatabaseTrace(
            query,
            dbapi2_module=PostgresApi,
            connect_params=getattr(self, "_nr_connect_params", None),
        ):
            return await self.__wrapped__.query(query, *args, **kwargs)

    async def prepare(self, stmt_name, query, *args, **kwargs):
        with DatabaseTrace(
            "PREPARE {stmt_name} FROM '{query}'".format(
                stmt_name=stmt_name, query=query
            ),
            dbapi2_module=PostgresApi,
            connect_params=getattr(self, "_nr_connect_params", None),
        ):
            return await self.__wrapped__.prepare(stmt_name, query, *args, **kwargs)

    async def copy_in(self, copy_stmt, *args, **kwargs):
        with DatabaseTrace(
            copy_stmt,
            dbapi2_module=PostgresApi,
            connect_params=getattr(self, "_nr_connect_params", None),
        ):
            return await self.__wrapped__.copy_in(copy_stmt, *args, **kwargs)

    async def copy_out(self, copy_stmt, *args, **kwargs):
        with DatabaseTrace(
            copy_stmt,
            dbapi2_module=PostgresApi,
            connect_params=getattr(self, "_nr_connect_params", None),
        ):
            return await self.__wrapped__.copy_out(copy_stmt, *args, **kwargs)


def proxy_protocol(wrapped, instance, args, kwargs):
    proxy = ProtocolProxy(wrapped(*args, **kwargs))
    proxy._nr_connect_params = (args, kwargs)
    return proxy


def wrap_connect(wrapped, instance, args, kwargs):
    host = port = database_name = None
    if "addr" in kwargs:
        host, port, database_name = PostgresApi._instance_info(
            kwargs["addr"], None, kwargs.get("params")
        )

    with DatastoreTrace(
        PostgresApi._nr_database_product,
        None,
        "connect",
        host=host,
        port_path_or_id=port,
        database_name=database_name,
    ):
        return wrapped(*args, **kwargs)


def instrument_asyncpg_protocol(module):
    wrap_function_wrapper(module, "Protocol", proxy_protocol)


def instrument_asyncpg_connect_utils(module):
    wrap_function_wrapper(module, "_connect_addr", wrap_connect)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from newrelic.api.time_trace import TimeTrace, current_trace
from newrelic.api.transaction import current_transaction
from newrelic.common.async_wrapper import async_wrapper
from newrelic.common.object_wrapper import FunctionWrapper, wrap_object
from newrelic.core.graphql_node import GraphQLOperationNode, GraphQLResolverNode


class GraphQLOperationTrace(TimeTrace):
    def __init__(self, **kwargs):
        parent = None
        if kwargs:
            if len(kwargs) > 1:
                raise TypeError("Invalid keyword arguments:", kwargs)
            parent = kwargs["parent"]
        super(GraphQLOperationTrace, self).__init__(parent)

        self.operation_name = "<anonymous>"
        self.operation_type = "<unknown>"
        self.deepest_path = "<unknown>"
        self.graphql = None
        self.graphql_format = None
        self.statement = None
        self.product = "GraphQL"

    def __repr__(self):
        return "<%s object at 0x%x %s>" % (
            self.__class__.__name__,
            id(self),
            dict(
                operation_name=self.operation_name,
                operation_type=self.operation_type,
                deepest_path=self.deepest_path,
                graphql=self.graphql,
            ),
        )

    @property
    def formatted(self):
        if not self.statement:
            return "<unknown>"

        transaction = current_transaction(active_only=False)

        # Record SQL settings
        settings = transaction.settings
        tt = settings.transaction_tracer
        self.graphql_format = tt.record_sql

        return self.statement.formatted(self.graphql_format)

    def finalize_data(self, transaction, exc=None, value=None, tb=None):
        # Add attributes
        self._add_agent_attribute("graphql.operation.type", self.operation_type)
        self._add_agent_attribute("graphql.operation.name", self.operation_name)

        # Attach formatted graphql
        limit = transaction.settings.agent_limits.sql_query_length_maximum
        self.graphql = graphql = self.formatted[:limit]
        self._add_agent_attribute("graphql.operation.query", graphql)

        return super(GraphQLOperationTrace, self).finalize_data(transaction, exc=None, value=None, tb=None)

    def create_node(self):
        return GraphQLOperationNode(
            children=self.children,
            start_time=self.start_time,
            end_time=self.end_time,
            duration=self.duration,
            exclusive=self.exclusive,
            guid=self.guid,
            agent_attributes=self.agent_attributes,
            user_attributes=self.user_attributes,
            operation_name=self.operation_name,
            operation_type=self.operation_type,
            deepest_path=self.deepest_path,
            graphql=self.graphql,
            product=self.product,
        )

    def set_transaction_name(self, priority=None):
        transaction = current_transaction()
        if transaction:
            name = (
                "%s/%s/%s" % (self.operation_type, self.operation_name, self.deepest_path)
                if self.deepest_path
                else "%s/%s" % (self.operation_type, self.operation_name)
            )
            transaction.set_transaction_name(name, "GraphQL", priority=priority)


def GraphQLOperationTraceWrapper(wrapped):
    def _nr_graphql_trace_wrapper_(wrapped, instance, args, kwargs):
        wrapper = async_wrapper(wrapped)
        if not wrapper:
            parent = current_trace()
            if not parent:
                return wrapped(*args, **kwargs)
        else:
            parent = None

        trace = GraphQLOperationTrace(parent=parent)

        if wrapper:  # pylint: disable=W0125,W0126
            return wrapper(wrapped, trace)(*args, **kwargs)

        with trace:
            return wrapped(*args, **kwargs)

    return FunctionWrapper(wrapped, _nr_graphql_trace_wrapper_)


def graphql_operation_trace():
    return functools.partial(GraphQLOperationTraceWrapper)


def wrap_graphql_operation_trace(module, object_path):
    wrap_object(module, object_path, GraphQLOperationTraceWrapper)


class GraphQLResolverTrace(TimeTrace):
    def __init__(self, field_name=None, **kwargs):
        super(GraphQLResolverTrace, self).__init__(**kwargs)
        self.field_name = field_name
        self._product = None

    def __repr__(self):
        return "<%s object at 0x%x %s>" % (self.__class__.__name__, id(self), dict(field_name=self.field_name))

    def __enter__(self):
        super(GraphQLResolverTrace, self).__enter__()
        _ = self.product  # Cache product value
        return self

    @property
    def product(self):
        if not self._product:
            # Find GraphQLOperationTrace to obtain stored product info
            parent = self  # init to self for loop start
            while parent is not None and not isinstance(parent, GraphQLOperationTrace):
                parent = getattr(parent, "parent", None)

            if parent is not None:
                self._product = getattr(parent, "product", "GraphQL")
            else:
                self._product = "GraphQL"

        return self._product

    def finalize_data(self, *args, **kwargs):
        self._add_agent_attribute("graphql.field.name", self.field_name)

        return super(GraphQLResolverTrace, self).finalize_data(*args, **kwargs)

    def create_node(self):
        return GraphQLResolverNode(
            field_name=self.field_name,
            children=self.children,
            start_time=self.start_time,
            end_time=self.end_time,
            duration=self.duration,
            exclusive=self.exclusive,
            guid=self.guid,
            agent_attributes=self.agent_attributes,
            user_attributes=self.user_attributes,
            product=self.product,
        )


def GraphQLResolverTraceWrapper(wrapped):
    def _nr_graphql_trace_wrapper_(wrapped, instance, args, kwargs):
        wrapper = async_wrapper(wrapped)
        if not wrapper:
            parent = current_trace()
            if not parent:
                return wrapped(*args, **kwargs)
        else:
            parent = None

        trace = GraphQLResolverTrace(parent=parent)

        if wrapper:  # pylint: disable=W0125,W0126
            return wrapper(wrapped, trace)(*args, **kwargs)

        with trace:
            return wrapped(*args, **kwargs)

    return FunctionWrapper(wrapped, _nr_graphql_trace_wrapper_)


def graphql_resolver_trace():
    return functools.partial(GraphQLResolverTraceWrapper)


def wrap_graphql_resolver_trace(module, object_path):
    wrap_object(module, object_path, GraphQLResolverTraceWrapper)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import threading

try:
    import grpc

    from newrelic.core.infinite_tracing_pb2 import RecordStatus, Span
except ImportError:
    grpc = None

_logger = logging.getLogger(__name__)


class StreamingRpc(object):
    """Streaming Remote Procedure Call

    This class keeps a stream_stream RPC alive, retrying after a timeout when
    errors are encountered. If grpc.StatusCode.UNIMPLEMENTED is encountered, a
    retry will not occur.
    """

    PATH = "/com.newrelic.trace.v1.IngestService/RecordSpan"
    RETRY_POLICY = (
        (15, False),
        (15, False),
        (30, False),
        (60, False),
        (120, False),
        (300, True),
    )
    OPTIONS = [("grpc.enable_retries", 0)]

    def __init__(self, endpoint, stream_buffer, metadata, record_metric, ssl=True):
        self._endpoint = endpoint
        self._ssl = ssl
        self.metadata = metadata
        self.stream_buffer = stream_buffer
        self.request_iterator = iter(stream_buffer)
        self.response_processing_thread = threading.Thread(
            target=self.process_responses, name="NR-StreamingRpc-process-responses"
        )
        self.response_processing_thread.daemon = True
        self.notify = self.condition()
        self.record_metric = record_metric
        self.closed = False

        self.create_channel()

    def create_channel(self):
        if self._ssl:
            credentials = grpc.ssl_channel_credentials()
            self.channel = grpc.secure_channel(self._endpoint, credentials, options=self.OPTIONS)
        else:
            self.channel = grpc.insecure_channel(self._endpoint, options=self.OPTIONS)

        self.rpc = self.channel.stream_stream(self.PATH, Span.SerializeToString, RecordStatus.FromString)

    def create_response_iterator(self):
        with self.stream_buffer._notify:
            self.request_iterator = iter(self.stream_buffer)
            self.request_iterator._stream = reponse_iterator = self.rpc(self.request_iterator, metadata=self.metadata)
            return reponse_iterator

    @staticmethod
    def condition(*args, **kwargs):
        return threading.Condition(*args, **kwargs)

    def close(self):
        channel = None
        with self.notify:
            if self.channel:
                channel = self.channel
                self.channel = None
                self.closed = True
            self.notify.notify_all()

        if channel:
            _logger.debug("Closing streaming rpc.")
            channel.close()
            try:
                self.response_processing_thread.join(timeout=5)
            except Exception:
                pass
            _logger.debug("Streaming rpc close completed.")

    def connect(self):
        self.response_processing_thread.start()

    def process_responses(self):
        response_iterator = None

        retry = 0
        while True:
            with self.notify:
                if self.channel and response_iterator:
                    code = response_iterator.code()
                    details = response_iterator.details()

                    self.record_metric(
                        "Supportability/InfiniteTracing/Span/gRPC/%s" % code.name,
                        {"count": 1},
                    )

                    if code is grpc.StatusCode.OK:
                        _logger.debug(
                            "Streaming RPC received OK "
                            "response code. The agent will attempt "
                            "to reestablish the stream immediately."
                        )

                        # Reconnect channel for load balancing
                        self.request_iterator.shutdown()
                        self.channel.close()
                        self.create_channel()

                    else:
                        self.record_metric(
                            "Supportability/InfiniteTracing/Span/Response/Error",
                            {"count": 1},
                        )

                        if code is grpc.StatusCode.UNIMPLEMENTED:
                            _logger.error(
                                "Streaming RPC received "
                                "UNIMPLEMENTED response code. "
                                "The agent will not attempt to "
                                "reestablish the stream."
                            )
                            break

                        # Unpack retry policy settings
                        if retry >= len(self.RETRY_POLICY):
                            retry_time, error = self.RETRY_POLICY[-1]
                        else:
                            retry_time, error = self.RETRY_POLICY[retry]
                            retry += 1

                        # Emit appropriate retry logs
                        if not error:
                            _logger.warning(
                                "Streaming RPC closed. Will attempt to reconnect in %d seconds. Check the prior log entries and remedy any issue as necessary, or if the problem persists, report this problem to New Relic support for further investigation. Code: %s Details: %s",
                                retry_time,
                                code,
                                details,
                            )
                        else:
                            _logger.error(
                                "Streaming RPC closed after additional attempts. Will attempt to reconnect in %d seconds. Please report this problem to New Relic support for further investigation. Code: %s Details: %s",
                                retry_time,
                                code,
                                details,
                            )

                        # Reconnect channel with backoff
                        self.request_iterator.shutdown()
                        self.channel.close()
                        self.notify.wait(retry_time)
                        if self.closed:
                            break
                        else:
                            _logger.debug("Attempting to reconnect Streaming RPC.")
                            self.create_channel()

                if self.closed:
                    break

                response_iterator = self.create_response_iterator()

                _logger.info("Streaming RPC connect completed.")

            try:
                for response in response_iterator:
                    _logger.debug("Stream response: %s", response)
            except Exception:
                pass

        self.close()
        _logger.info("Process response thread ending.")



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from newrelic.api.cat_header_mixin import CatHeaderMixin
from newrelic.api.time_trace import TimeTrace, current_trace
from newrelic.common.async_wrapper import async_wrapper
from newrelic.common.object_wrapper import FunctionWrapper, wrap_object
from newrelic.core.message_node import MessageNode


class MessageTrace(CatHeaderMixin, TimeTrace):

    cat_id_key = "NewRelicID"
    cat_transaction_key = "NewRelicTransaction"
    cat_appdata_key = "NewRelicAppData"
    cat_synthetics_key = "NewRelicSynthetics"

    def __init__(self, library, operation, destination_type, destination_name, params=None, **kwargs):

        parent = None
        if kwargs:
            if len(kwargs) > 1:
                raise TypeError("Invalid keyword arguments:", kwargs)
            parent = kwargs["parent"]
        super(MessageTrace, self).__init__(parent)

        self.library = library
        self.operation = operation

        self.params = params

        self.destination_type = destination_type
        self.destination_name = destination_name

    def __enter__(self):
        result = super(MessageTrace, self).__enter__()

        if result and self.transaction:
            self.library = self.transaction._intern_string(self.library)
            self.operation = self.transaction._intern_string(self.operation)

        # Only record parameters when not high security mode and only
        # when enabled in settings.
        if not (
            self.should_record_segment_params
            and self.settings
            and self.settings.message_tracer.segment_parameters_enabled
        ):
            self.params = None
        return result

    def __repr__(self):
        return "<%s object at 0x%x %s>" % (
            self.__class__.__name__,
            id(self),
            dict(library=self.library, operation=self.operation),
        )

    def terminal_node(self):
        return True

    def create_node(self):
        return MessageNode(
            library=self.library,
            operation=self.operation,
            children=self.children,
            start_time=self.start_time,
            end_time=self.end_time,
            duration=self.duration,
            exclusive=self.exclusive,
            destination_name=self.destination_name,
            destination_type=self.destination_type,
            params=self.params,
            guid=self.guid,
            agent_attributes=self.agent_attributes,
            user_attributes=self.user_attributes,
        )


def MessageTraceWrapper(wrapped, library, operation, destination_type, destination_name, params={}):
    def _nr_message_trace_wrapper_(wrapped, instance, args, kwargs):
        wrapper = async_wrapper(wrapped)
        if not wrapper:
            parent = current_trace()
            if not parent:
                return wrapped(*args, **kwargs)
        else:
            parent = None

        if callable(library):
            if instance is not None:
                _library = library(instance, *args, **kwargs)
            else:
                _library = library(*args, **kwargs)
        else:
            _library = library

        if callable(operation):
            if instance is not None:
                _operation = operation(instance, *args, **kwargs)
            else:
                _operation = operation(*args, **kwargs)
        else:
            _operation = operation

        if callable(destination_type):
            if instance is not None:
                _destination_type = destination_type(instance, *args, **kwargs)
            else:
                _destination_type = destination_type(*args, **kwargs)
        else:
            _destination_type = destination_type

        if callable(destination_name):
            if instance is not None:
                _destination_name = destination_name(instance, *args, **kwargs)
            else:
                _destination_name = destination_name(*args, **kwargs)
        else:
            _destination_name = destination_name

        trace = MessageTrace(_library, _operation, _destination_type, _destination_name, params={}, parent=parent)

        if wrapper:  # pylint: disable=W0125,W0126
            return wrapper(wrapped, trace)(*args, **kwargs)

        with trace:
            return wrapped(*args, **kwargs)

    return FunctionWrapper(wrapped, _nr_message_trace_wrapper_)


def message_trace(library, operation, destination_type, destination_name, params={}):
    return functools.partial(
        MessageTraceWrapper,
        library=library,
        operation=operation,
        destination_type=destination_type,
        destination_name=destination_name,
        params=params,
    )


def wrap_message_trace(module, object_path, library, operation, destination_type, destination_name, params={}):
    wrap_object(
        module, object_path, MessageTraceWrapper, (library, operation, destination_type, destination_name, params)
    )



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.asgi_application import ASGIApplicationWrapper


@property
def loaded_app(self):
    # Always use the original application until the interface is resolved in
    # auto mode
    if getattr(self, "interface", "") == "auto":
        app = self._nr_loaded_app
        while hasattr(app, "__wrapped__"):
            app = app.__wrapped__
        return app
    return self._nr_loaded_app


@loaded_app.setter
def loaded_app(self, value):
    # Wrap only the first loaded app
    if (
        not getattr(self, "_nr_loaded_app", None)
        and value
        and getattr(self, "interface", "") != "wsgi"
    ):
        value = ASGIApplicationWrapper(value)
    self._nr_loaded_app = value


def instrument_uvicorn_config(module):
    module.Config.loaded_app = loaded_app



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from newrelic.api.datastore_trace import DatastoreTrace
from newrelic.api.transaction import current_transaction
from newrelic.common.object_wrapper import wrap_function_wrapper
from newrelic.hooks.datastore_redis import _conn_attrs_to_dict, _instance_info, _redis_client_methods, _redis_multipart_commands, _redis_operation_re


def _wrap_Aredis_method_wrapper_(module, instance_class_name, operation):
    async def _nr_wrapper_Aredis_method_(wrapped, instance, args, kwargs):
        transaction = current_transaction()
        if transaction is None:
            return await wrapped(*args, **kwargs)

        dt = DatastoreTrace(product="Redis", target=None, operation=operation)

        transaction._nr_datastore_instance_info = (None, None, None)

        with dt:
            result = await wrapped(*args, **kwargs)

            host, port_path_or_id, db = transaction._nr_datastore_instance_info
            dt.host = host
            dt.port_path_or_id = port_path_or_id
            dt.database_name = db
            return result

    name = "%s.%s" % (instance_class_name, operation)
    wrap_function_wrapper(module, name, _nr_wrapper_Aredis_method_)


def instrument_aredis_client(module):
    if hasattr(module, "StrictRedis"):
        for name in _redis_client_methods:
            if hasattr(module.StrictRedis, name):
                _wrap_Aredis_method_wrapper_(module, "StrictRedis", name)


async def _nr_Connection_send_command_wrapper_(wrapped, instance, args, kwargs):
    transaction = current_transaction()
    if not transaction:
        return await wrapped(*args, **kwargs)
        
    host, port_path_or_id, db = (None, None, None)

    try:
        dt = transaction.settings.datastore_tracer
        if dt.instance_reporting.enabled or dt.database_name_reporting.enabled:
            conn_kwargs = _conn_attrs_to_dict(instance)
            host, port_path_or_id, db = _instance_info(conn_kwargs)
    except Exception:
        pass

    transaction._nr_datastore_instance_info = (host, port_path_or_id, db)
        
    # Older Redis clients would when sending multi part commands pass
    # them in as separate arguments to send_command(). Need to therefore
    # detect those and grab the next argument from the set of arguments.

    operation = args[0].strip().lower()

    # If it's not a multi part command, there's no need to trace it, so
    # we can return early.

    if operation.split()[0] not in _redis_multipart_commands:
        return await wrapped(*args, **kwargs)

    # Convert multi args to single arg string

    if operation in _redis_multipart_commands and len(args) > 1:
        operation = "%s %s" % (operation, args[1].strip().lower())

    operation = _redis_operation_re.sub("_", operation)

    with DatastoreTrace(
        product="Redis", target=None, operation=operation, host=host, port_path_or_id=port_path_or_id, database_name=db
    ):
        return await wrapped(*args, **kwargs)
    

def instrument_aredis_connection(module):
    wrap_function_wrapper(module.Connection, "send_command", _nr_Connection_send_command_wrapper_)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import logging
import os
import random
import re
import sys
import threading
import time
import warnings
import weakref
from collections import OrderedDict

import newrelic.core.database_node
import newrelic.core.error_node
import newrelic.core.root_node
import newrelic.core.transaction_node
import newrelic.packages.six as six
from newrelic.api.time_trace import TimeTrace
from newrelic.common.encoding_utils import (
    DistributedTracePayload,
    NrTraceState,
    W3CTraceParent,
    W3CTraceState,
    base64_decode,
    convert_to_cat_metadata_value,
    deobfuscate,
    ensure_str,
    generate_path_hash,
    json_decode,
    json_encode,
    obfuscate,
)
from newrelic.core.attribute import (
    MAX_NUM_USER_ATTRIBUTES,
    create_agent_attributes,
    create_attributes,
    create_user_attributes,
    process_user_attribute,
)
from newrelic.core.attribute_filter import (
    DST_ERROR_COLLECTOR,
    DST_NONE,
    DST_TRANSACTION_TRACER,
)
from newrelic.core.config import DEFAULT_RESERVOIR_SIZE
from newrelic.core.custom_event import create_custom_event
from newrelic.core.stack_trace import exception_stack
from newrelic.core.stats_engine import CustomMetrics, SampledDataSet
from newrelic.core.thread_utilization import utilization_tracker
from newrelic.core.trace_cache import (
    TraceCacheActiveTraceError,
    TraceCacheNoActiveTraceError,
    trace_cache,
)

_logger = logging.getLogger(__name__)

DISTRIBUTED_TRACE_KEYS_REQUIRED = ("ty", "ac", "ap", "tr", "ti")
DISTRIBUTED_TRACE_TRANSPORT_TYPES = set(("HTTP", "HTTPS", "Kafka", "JMS", "IronMQ", "AMQP", "Queue", "Other"))
DELIMITER_FORMAT_RE = re.compile("[ \t]*,[ \t]*")
ACCEPTED_DISTRIBUTED_TRACE = 1
CREATED_DISTRIBUTED_TRACE = 2
PARENT_TYPE = {
    "0": "App",
    "1": "Browser",
    "2": "Mobile",
}


class Sentinel(TimeTrace):
    def __init__(self, transaction):
        super(Sentinel, self).__init__(None)
        self.transaction = transaction

        # Set the thread id to the same as the transaction
        self.thread_id = transaction.thread_id

    def add_child(self, node):
        self.children.append(node)

    def update_with_transaction_custom_attributes(self, transaction_params):
        """
        Loops through the transaction attributes and adds them to the
        root span's user attributes.
        """
        for key, value in transaction_params.items():
            if len(self.user_attributes) >= MAX_NUM_USER_ATTRIBUTES:
                _logger.debug(
                    "Maximum number of custom attributes already "
                    "added to span. Some transaction attributes may "
                    "not be included."
                )
                break
            if key not in self.user_attributes:
                self.user_attributes[key] = value

    def complete_root(self):
        try:
            trace_cache().complete_root(self)
        finally:
            self.exited = True

    @staticmethod
    def complete_trace():
        pass

    @property
    def transaction(self):
        return self._transaction and self._transaction()

    @transaction.setter
    def transaction(self, value):
        if value:
            self._transaction = weakref.ref(value)

    @property
    def root(self):
        return self

    @root.setter
    def root(self, value):
        pass


class CachedPath(object):
    def __init__(self, transaction):
        self._name = None
        self.transaction = weakref.ref(transaction)

    def path(self):
        if self._name is not None:
            return self._name

        transaction = self.transaction()
        if transaction:
            return transaction.path

        return "Unknown"


class Transaction(object):

    STATE_PENDING = 0
    STATE_RUNNING = 1
    STATE_STOPPED = 2

    def __init__(self, application, enabled=None):

        self._application = application

        self.thread_id = None

        self._transaction_id = id(self)
        self._transaction_lock = threading.Lock()

        self._dead = False

        self._state = self.STATE_PENDING
        self._settings = None

        self._name_priority = 0
        self._group = None
        self._name = None
        self._cached_path = CachedPath(self)
        self._loop_time = 0.0

        self._frameworks = set()

        self._frozen_path = None

        self.root_span = None

        self._request_uri = None
        self._port = None

        self.queue_start = 0.0

        self.start_time = 0.0
        self.end_time = 0.0
        self.last_byte_time = 0.0

        self.total_time = 0.0

        self.stopped = False

        self._trace_node_count = 0

        self._errors = []
        self._slow_sql = []
        self._custom_events = SampledDataSet(capacity=DEFAULT_RESERVOIR_SIZE)

        self._stack_trace_count = 0
        self._explain_plan_count = 0

        self._string_cache = {}

        self._custom_params = OrderedDict()
        self._request_params = {}

        self._utilization_tracker = None

        self._thread_utilization_start = None
        self._thread_utilization_end = None
        self._thread_utilization_value = None

        self._cpu_user_time_start = None
        self._cpu_user_time_end = None
        self._cpu_user_time_value = 0.0

        self._read_length = None

        self._read_start = None
        self._read_end = None

        self._sent_start = None
        self._sent_end = None

        self._bytes_read = 0
        self._bytes_sent = 0

        self._calls_read = 0
        self._calls_readline = 0
        self._calls_readlines = 0

        self._calls_write = 0
        self._calls_yield = 0

        self._transaction_metrics = {}

        self._agent_attributes = {}

        self.background_task = False

        self.enabled = False
        self.autorum_disabled = False

        self.ignore_transaction = False
        self.suppress_apdex = False
        self.suppress_transaction_trace = False

        self.capture_params = None

        self.apdex = 0

        self.rum_token = None

        trace_id = "%032x" % random.getrandbits(128)

        # 16-digit random hex. Padded with zeros in the front.
        self.guid = trace_id[:16]

        # 32-digit random hex. Padded with zeros in the front.
        self._trace_id = trace_id

        # This may be overridden by processing an inbound CAT header
        self.parent_type = None
        self.parent_span = None
        self.trusted_parent_span = None
        self.tracing_vendors = None
        self.parent_tx = None
        self.parent_app = None
        self.parent_account = None
        self.parent_transport_type = None
        self.parent_transport_duration = None
        self.tracestate = ""
        self._priority = None
        self._sampled = None

        self._distributed_trace_state = 0

        self.client_cross_process_id = None
        self.client_account_id = None
        self.client_application_id = None
        self.referring_transaction_guid = None
        self.record_tt = False
        self._trip_id = None
        self._referring_path_hash = None
        self._alternate_path_hashes = {}
        self.is_part_of_cat = False

        self.synthetics_resource_id = None
        self.synthetics_job_id = None
        self.synthetics_monitor_id = None
        self.synthetics_header = None

        self._custom_metrics = CustomMetrics()

        global_settings = application.global_settings

        if global_settings.enabled:
            if enabled or (enabled is None and application.enabled):
                self._settings = application.settings
                if not self._settings:
                    application.activate()

                    # We see again if the settings is now valid
                    # in case startup timeout had been specified
                    # and registration had been started and
                    # completed within the timeout.

                    self._settings = application.settings

                if self._settings:
                    self.enabled = True

    def __del__(self):
        self._dead = True
        if self._state == self.STATE_RUNNING:
            self.__exit__(None, None, None)

    def __enter__(self):

        assert self._state == self.STATE_PENDING

        # Bail out if the transaction is not enabled.

        if not self.enabled:
            return self

        # Record the start time for transaction.

        self.start_time = time.time()

        # Record initial CPU user time.

        self._cpu_user_time_start = os.times()[0]

        # Set the thread ID upon entering the transaction.
        # This is done here so that any asyncio tasks will
        # be active and the task ID will be used to
        # store traces into the trace cache.
        self.thread_id = trace_cache().current_thread_id()

        # Create the root span then push it
        # into the trace cache as the active trace.
        # If there is an active transaction already
        # mark this one as disabled and do not raise the
        # exception.
        self.root_span = root_span = Sentinel(self)

        try:
            trace_cache().save_trace(root_span)
        except TraceCacheActiveTraceError:
            self.enabled = False
            return self

        # Calculate initial thread utilisation factor.
        # For now we only do this if we know it is an
        # actual thread and not a greenlet.

        if not hasattr(sys, "_current_frames") or self.thread_id in sys._current_frames():
            try:
                thread_instance = threading.current_thread()
            except TypeError:
                thread_instance = threading.currentThread()

            self._utilization_tracker = utilization_tracker(self.application.name)
            if self._utilization_tracker:
                self._utilization_tracker.enter_transaction(thread_instance)
                self._thread_utilization_start = self._utilization_tracker.utilization_count()

        # Mark transaction as active and update state
        # used to validate correct usage of class.

        self._state = self.STATE_RUNNING

        return self

    def __exit__(self, exc, value, tb):

        # Bail out if the transaction is not enabled.

        if not self.enabled:
            return

        if self._transaction_id != id(self):
            return

        if not self._settings:
            return

        # Record error if one was registered.

        if exc is not None and value is not None and tb is not None:
            self.root_span.notice_error((exc, value, tb))

        self._state = self.STATE_STOPPED

        # Force the root span out of the cache if it's there
        # This also prevents saving of the root span in the future since the
        # transaction will be None
        root = self.root_span
        try:
            root.complete_root()
        except TraceCacheNoActiveTraceError:
            # It's possible that the weakref can be cleared prior to a
            # finalizer call. This results in traces being implicitly dropped
            # from the cache even though they're still referenced at this time.
            #
            # https://bugs.python.org/issue40312
            if not self._dead:
                _logger.exception(
                    "Runtime instrumentation error. Attempt to "
                    "drop the trace but where none is active. "
                    "Report this issue to New Relic support."
                ),
                return
        except Exception:
            _logger.exception(
                "Runtime instrumentation error. Exception "
                "occurred during exit. Report this issue to New Relic support."
            )
            return

        # Record the end time for transaction and then
        # calculate the duration.

        if not self.stopped:
            self.end_time = time.time()

        # Calculate transaction duration

        duration = self.end_time - self.start_time

        # Calculate response time. Calculation depends on whether
        # a web response was sent back.

        if self.last_byte_time == 0.0:
            response_time = duration
        else:
            response_time = self.last_byte_time - self.start_time

        # Calculate overall user time.

        if not self._cpu_user_time_end:
            self._cpu_user_time_end = os.times()[0]

        if duration and self._cpu_user_time_end:
            self._cpu_user_time_value = self._cpu_user_time_end - self._cpu_user_time_start

        # Calculate thread utilisation factor. Note that even if
        # we are tracking thread utilization we skip calculation
        # if duration is zero. Under normal circumstances this
        # should not occur but may if the system clock is wound
        # backwards and duration was squashed to zero due to the
        # request appearing to finish before it started. It may
        # also occur if true response time came in under the
        # resolution of the clock being used, but that is highly
        # unlikely as the overhead of the agent itself should
        # always ensure that that is hard to achieve.

        if self._utilization_tracker:
            self._utilization_tracker.exit_transaction()
            if self._thread_utilization_start is not None and duration > 0.0:
                if not self._thread_utilization_end:
                    self._thread_utilization_end = self._utilization_tracker.utilization_count()
                self._thread_utilization_value = (
                    self._thread_utilization_end - self._thread_utilization_start
                ) / duration

        self._freeze_path()

        # _sent_end should already be set by this point, but in case it
        # isn't, set it now before we record the custom metrics and derive
        # agent attributes

        if self._sent_start:
            if not self._sent_end:
                self._sent_end = time.time()

        request_params = self.request_parameters

        root.update_with_transaction_custom_attributes(self._custom_params)

        # Update agent attributes and include them on the root node
        self._update_agent_attributes()
        root_agent_attributes = dict(self._agent_attributes)
        root_agent_attributes.update(request_params)
        root_agent_attributes.update(root.agent_attributes)

        exclusive = duration + root.exclusive

        root_node = newrelic.core.root_node.RootNode(
            name=self.name_for_metric,
            children=tuple(root.children),
            start_time=self.start_time,
            end_time=self.end_time,
            exclusive=exclusive,
            duration=duration,
            guid=root.guid,
            agent_attributes=root_agent_attributes,
            user_attributes=root.user_attributes,
            path=self.path,
            trusted_parent_span=self.trusted_parent_span,
            tracing_vendors=self.tracing_vendors,
        )

        # Add transaction exclusive time to total exclusive time
        #
        self.total_time += exclusive

        if self.client_cross_process_id is not None:
            metric_name = "ClientApplication/%s/all" % (self.client_cross_process_id)
            self.record_custom_metric(metric_name, duration)

        # Record supportability metrics for api calls

        for key, value in six.iteritems(self._transaction_metrics):
            self.record_custom_metric(key, {"count": value})

        if self._frameworks:
            for framework, version in self._frameworks:
                self.record_custom_metric("Python/Framework/%s/%s" % (framework, version), 1)

        if self._settings.distributed_tracing.enabled:
            # Sampled and priority need to be computed at the end of the
            # transaction when distributed tracing or span events are enabled.
            self._compute_sampled_and_priority()

        self._cached_path._name = self.path
        agent_attributes = self.agent_attributes
        agent_attributes.extend(self.filter_request_parameters(request_params))
        node = newrelic.core.transaction_node.TransactionNode(
            settings=self._settings,
            path=self.path,
            type=self.type,
            group=self.group_for_metric,
            base_name=self._name,
            name_for_metric=self.name_for_metric,
            port=self._port,
            request_uri=self._request_uri,
            queue_start=self.queue_start,
            start_time=self.start_time,
            end_time=self.end_time,
            last_byte_time=self.last_byte_time,
            total_time=self.total_time,
            response_time=response_time,
            duration=duration,
            exclusive=exclusive,
            errors=tuple(self._errors),
            slow_sql=tuple(self._slow_sql),
            custom_events=self._custom_events,
            apdex_t=self.apdex,
            suppress_apdex=self.suppress_apdex,
            custom_metrics=self._custom_metrics,
            guid=self.guid,
            cpu_time=self._cpu_user_time_value,
            suppress_transaction_trace=self.suppress_transaction_trace,
            client_cross_process_id=self.client_cross_process_id,
            referring_transaction_guid=self.referring_transaction_guid,
            record_tt=self.record_tt,
            synthetics_resource_id=self.synthetics_resource_id,
            synthetics_job_id=self.synthetics_job_id,
            synthetics_monitor_id=self.synthetics_monitor_id,
            synthetics_header=self.synthetics_header,
            is_part_of_cat=self.is_part_of_cat,
            trip_id=self.trip_id,
            path_hash=self.path_hash,
            referring_path_hash=self._referring_path_hash,
            alternate_path_hashes=self.alternate_path_hashes,
            trace_intrinsics=self.trace_intrinsics,
            distributed_trace_intrinsics=self.distributed_trace_intrinsics,
            agent_attributes=agent_attributes,
            user_attributes=self.user_attributes,
            priority=self.priority,
            sampled=self.sampled,
            parent_span=self.parent_span,
            parent_transport_duration=self.parent_transport_duration,
            parent_type=self.parent_type,
            parent_account=self.parent_account,
            parent_app=self.parent_app,
            parent_tx=self.parent_tx,
            parent_transport_type=self.parent_transport_type,
            root_span_guid=root.guid,
            trace_id=self.trace_id,
            loop_time=self._loop_time,
            root=root_node,
        )

        # Clear settings as we are all done and don't need it
        # anymore.

        self._settings = None
        self.enabled = False

        # Unless we are ignoring the transaction, record it. We
        # need to lock the profile samples and replace it with
        # an empty list just in case the thread profiler kicks
        # in just as we are trying to record the transaction.
        # If we don't, when processing the samples, addition of
        # new samples can cause an error.

        if not self.ignore_transaction:

            self._application.record_transaction(node)

    @property
    def sampled(self):
        return self._sampled

    @property
    def priority(self):
        return self._priority

    @property
    def state(self):
        return self._state

    @property
    def is_distributed_trace(self):
        return self._distributed_trace_state != 0

    @property
    def settings(self):
        return self._settings

    @property
    def application(self):
        return self._application

    @property
    def type(self):
        if self.background_task:
            transaction_type = "OtherTransaction"
        else:
            transaction_type = "WebTransaction"
        return transaction_type

    @property
    def name(self):
        return self._name

    @property
    def group(self):
        return self._group

    @property
    def name_for_metric(self):
        """Combine group and name for use as transaction name in metrics."""
        group = self.group_for_metric

        transaction_name = self._name

        if transaction_name is None:
            transaction_name = "<undefined>"

        # Stripping the leading slash on the request URL held by
        # transaction_name when type is 'Uri' is to keep compatibility
        # with PHP agent and also possibly other agents. Leading
        # slash it not deleted for other category groups as the
        # leading slash may be significant in that situation.

        if group in ("Uri", "NormalizedUri") and transaction_name.startswith("/"):
            name = "%s%s" % (group, transaction_name)
        else:
            name = "%s/%s" % (group, transaction_name)

        return name

    @property
    def group_for_metric(self):
        _group = self._group

        if _group is None:
            if self.background_task:
                _group = "Python"
            else:
                _group = "Uri"

        return _group

    @property
    def path(self):
        if self._frozen_path:
            return self._frozen_path

        return "%s/%s" % (self.type, self.name_for_metric)

    @property
    def trip_id(self):
        return self._trip_id or self.guid

    @property
    def trace_id(self):
        return self._trace_id

    @property
    def alternate_path_hashes(self):
        """Return the alternate path hashes but not including the current path
        hash.

        """
        return sorted(set(self._alternate_path_hashes.values()) - set([self.path_hash]))

    @property
    def path_hash(self):
        """Path hash is a 32-bit digest of the string "appname;txn_name"
        XORed with the referring_path_hash. Since the txn_name can change
        during the course of a transaction, up to 10 path_hashes are stored
        in _alternate_path_hashes. Before generating the path hash, check the
        _alternate_path_hashes to determine if we've seen this identifier and
        return the value.

        """

        if not self.is_part_of_cat:
            return None

        identifier = "%s;%s" % (self.application.name, self.path)

        # Check if identifier is already part of the _alternate_path_hashes and
        # return the value if available.

        if self._alternate_path_hashes.get(identifier):
            return self._alternate_path_hashes[identifier]

        # If the referring_path_hash is unavailable then we use '0' as the
        # seed.

        try:
            seed = int((self._referring_path_hash or "0"), base=16)
        except Exception:
            seed = 0

        try:
            path_hash = generate_path_hash(identifier, seed)
        except ValueError:
            _logger.warning(
                "Unable to generate cross application tracer headers. "
                "MD5 hashing may not be available. (Is this system FIPS compliant?) "
                "We recommend enabling distributed tracing instead. For details and a transition guide see "
                "https://docs.newrelic.com/docs/agents/python-agent/configuration/python-agent-configuration#distributed-tracing-settings"
            )
            return None

        # Only store up to 10 alternate path hashes.

        if len(self._alternate_path_hashes) < 10:
            self._alternate_path_hashes[identifier] = path_hash

        return path_hash

    @property
    def attribute_filter(self):
        return self._settings.attribute_filter

    @property
    def read_duration(self):
        read_duration = 0
        if self._read_start and self._read_end:
            read_duration = self._read_end - self._read_start
        return read_duration

    @property
    def sent_duration(self):
        sent_duration = 0
        if self._sent_start and self._sent_end:
            sent_duration = self._sent_end - self._sent_start
        return sent_duration

    @property
    def queue_wait(self):
        queue_wait = 0
        if self.queue_start:
            queue_wait = self.start_time - self.queue_start
            if queue_wait < 0:
                queue_wait = 0
        return queue_wait

    @property
    def should_record_segment_params(self):
        # Only record parameters when it is safe to do so
        return self.settings and not self.settings.high_security

    @property
    def trace_intrinsics(self):
        """Intrinsic attributes for transaction traces and error traces"""
        i_attrs = {}

        if self.referring_transaction_guid:
            i_attrs["referring_transaction_guid"] = self.referring_transaction_guid
        if self.client_cross_process_id:
            i_attrs["client_cross_process_id"] = self.client_cross_process_id
        if self.trip_id:
            i_attrs["trip_id"] = self.trip_id
        if self.path_hash:
            i_attrs["path_hash"] = self.path_hash
        if self.synthetics_resource_id:
            i_attrs["synthetics_resource_id"] = self.synthetics_resource_id
        if self.synthetics_job_id:
            i_attrs["synthetics_job_id"] = self.synthetics_job_id
        if self.synthetics_monitor_id:
            i_attrs["synthetics_monitor_id"] = self.synthetics_monitor_id
        if self.total_time:
            i_attrs["totalTime"] = self.total_time
        if self._loop_time:
            i_attrs["eventLoopTime"] = self._loop_time

        # Add in special CPU time value for UI to display CPU burn.

        # XXX Disable cpu time value for CPU burn as was
        # previously reporting incorrect value and we need to
        # fix it, at least on Linux to report just the CPU time
        # for the executing thread.

        # if self._cpu_user_time_value:
        #     i_attrs['cpu_time'] = self._cpu_user_time_value

        i_attrs.update(self.distributed_trace_intrinsics)

        return i_attrs

    @property
    def distributed_trace_intrinsics(self):
        i_attrs = {}

        if not self._settings.distributed_tracing.enabled:
            return i_attrs

        i_attrs["guid"] = self.guid
        i_attrs["sampled"] = self.sampled
        i_attrs["priority"] = self.priority
        i_attrs["traceId"] = self.trace_id

        if not self._distributed_trace_state:
            return i_attrs

        if self.parent_type:
            i_attrs["parent.type"] = self.parent_type
        if self.parent_account:
            i_attrs["parent.account"] = self.parent_account
        if self.parent_app:
            i_attrs["parent.app"] = self.parent_app
        if self.parent_transport_type:
            i_attrs["parent.transportType"] = self.parent_transport_type
        if self.parent_transport_duration:
            i_attrs["parent.transportDuration"] = self.parent_transport_duration
        if self.trusted_parent_span:
            i_attrs["trustedParentId"] = self.trusted_parent_span
        if self.tracing_vendors:
            i_attrs["tracingVendors"] = self.tracing_vendors

        return i_attrs

    def filter_request_parameters(self, params):
        # Request parameters are a special case of agent attributes, so
        # they must be filtered separately

        # There are 3 cases we need to handle:
        #
        # 1. LEGACY: capture_params = False
        #
        #    Don't add request parameters at all, which means they will not
        #    go through the AttributeFilter.
        #
        # 2. LEGACY: capture_params = True
        #
        #    Filter request parameters through the AttributeFilter, but
        #    set the destinations to `TRANSACTION_TRACER | ERROR_COLLECTOR`.
        #
        #    If the user does not add any additional attribute filtering
        #    rules, this will result in the same outcome as the old
        #    capture_params = True behavior. They will be added to transaction
        #    traces and error traces.
        #
        # 3. CURRENT: capture_params is None
        #
        #    Filter request parameters through the AttributeFilter, but set
        #    the destinations to NONE.
        #
        #    That means by default, request parameters won't get included in
        #    any destination. But, it will allow user added include/exclude
        #    attribute filtering rules to be applied to the request parameters.
        attributes_request = []

        if self.capture_params is None:
            attributes_request = create_attributes(params, DST_NONE, self.attribute_filter)
        elif self.capture_params:
            attributes_request = create_attributes(
                params, DST_ERROR_COLLECTOR | DST_TRANSACTION_TRACER, self.attribute_filter
            )
        return attributes_request

    @property
    def request_parameters(self):
        if (self.capture_params is None) or self.capture_params:

            if self._request_params:

                r_attrs = {}

                for k, v in self._request_params.items():
                    new_key = "request.parameters.%s" % k
                    new_val = ",".join(v)

                    final_key, final_val = process_user_attribute(new_key, new_val)

                    if final_key:
                        r_attrs[final_key] = final_val

                return r_attrs
        return {}

    def _add_agent_attribute(self, key, value):
        self._agent_attributes[key] = value

    @property
    def agent_attributes(self):
        agent_attributes = create_agent_attributes(self._agent_attributes, self.attribute_filter)
        return agent_attributes

    def _update_agent_attributes(self):
        a_attrs = self._agent_attributes

        if self._settings.process_host.display_name:
            a_attrs["host.displayName"] = self._settings.process_host.display_name
        if self._thread_utilization_value:
            a_attrs["thread.concurrency"] = self._thread_utilization_value
        if self.queue_wait != 0:
            a_attrs["webfrontend.queue.seconds"] = self.queue_wait

    @property
    def user_attributes(self):
        return create_user_attributes(self._custom_params, self.attribute_filter)

    def _compute_sampled_and_priority(self):
        if self._priority is None:
            # truncate priority field to 6 digits past the decimal
            self._priority = float("%.6f" % random.random())  # nosec

        if self._sampled is None:
            self._sampled = self._application.compute_sampled()
            if self._sampled:
                self._priority += 1

    def _freeze_path(self):
        if self._frozen_path is None:
            self._name_priority = None

            if self._group == "Uri" and self._name != "/":
                # Apply URL normalization rules. We would only have raw
                # URLs where we were not specifically naming the web
                # transactions for a specific web framework to be a code
                # handler or otherwise.

                name, ignore = self._application.normalize_name(self._name, "url")

                if self._name != name:
                    self._group = "NormalizedUri"
                    self._name = name

                self.ignore_transaction = self.ignore_transaction or ignore

            # Apply transaction rules on the full transaction name.

            path, ignore = self._application.normalize_name(self.path, "transaction")

            self.ignore_transaction = self.ignore_transaction or ignore

            # Apply segment whitelist rule to the segments on the full
            # transaction name. The path is frozen at this point and cannot be
            # further changed.

            self._frozen_path, ignore = self._application.normalize_name(path, "segment")

            self.ignore_transaction = self.ignore_transaction or ignore

            # Look up the apdex from the table of key transactions. If
            # current transaction is not a key transaction then use the
            # default apdex from settings. The path used at this point
            # is the frozen path.

            self.apdex = self._settings.web_transactions_apdex.get(self.path) or self._settings.apdex_t

    def _record_supportability(self, metric_name):
        m = self._transaction_metrics.get(metric_name, 0)
        self._transaction_metrics[metric_name] = m + 1

    def _create_distributed_trace_data_with_guid(self, guid):
        data = self._create_distributed_trace_data()
        if guid and data and "id" in data:
            data["id"] = guid
        return data

    def _create_distributed_trace_data(self):
        if not self.enabled:
            return

        settings = self._settings
        account_id = settings.account_id
        trusted_account_key = settings.trusted_account_key
        application_id = settings.primary_application_id

        if not (account_id and application_id and trusted_account_key and settings.distributed_tracing.enabled):
            return

        self._compute_sampled_and_priority()
        data = dict(
            ty="App",
            ac=account_id,
            ap=application_id,
            tr=self.trace_id,
            sa=self.sampled,
            pr=self.priority,
            tx=self.guid,
            ti=int(time.time() * 1000.0),
        )

        if account_id != trusted_account_key:
            data["tk"] = trusted_account_key

        current_span = trace_cache().current_trace()
        if settings.span_events.enabled and settings.collect_span_events and current_span:
            data["id"] = current_span.guid

        self._distributed_trace_state |= CREATED_DISTRIBUTED_TRACE

        return data

    def _create_distributed_trace_payload(self):
        try:
            data = self._create_distributed_trace_data()
            if data is None:
                return
            payload = DistributedTracePayload(
                v=DistributedTracePayload.version,
                d=data,
            )
        except:
            self._record_supportability("Supportability/DistributedTrace/CreatePayload/Exception")
        else:
            self._record_supportability("Supportability/DistributedTrace/CreatePayload/Success")
            return payload

    def create_distributed_trace_payload(self):
        warnings.warn(
            (
                "The create_distributed_trace_payload API has been deprecated. "
                "Please use the insert_distributed_trace_headers API."
            ),
            DeprecationWarning,
        )
        return self._create_distributed_trace_payload()

    def _generate_distributed_trace_headers(self, data=None):
        try:
            data = data or self._create_distributed_trace_data()
            if data:

                traceparent = W3CTraceParent(data).text()
                yield ("traceparent", traceparent)

                tracestate = NrTraceState(data).text()
                if self.tracestate:
                    tracestate += "," + self.tracestate
                yield ("tracestate", tracestate)

                self._record_supportability("Supportability/TraceContext/Create/Success")

                if not self._settings.distributed_tracing.exclude_newrelic_header:
                    # Insert New Relic dt headers for backwards compatibility
                    payload = DistributedTracePayload(
                        v=DistributedTracePayload.version,
                        d=data,
                    )
                    yield ("newrelic", payload.http_safe())
                    self._record_supportability("Supportability/DistributedTrace/CreatePayload/Success")

        except:
            self._record_supportability("Supportability/TraceContext/Create/Exception")

            if not self._settings.distributed_tracing.exclude_newrelic_header:
                self._record_supportability("Supportability/DistributedTrace/CreatePayload/Exception")

    def insert_distributed_trace_headers(self, headers):
        headers.extend(self._generate_distributed_trace_headers())

    def _can_accept_distributed_trace_headers(self):
        if not self.enabled:
            return False

        settings = self._settings
        if not (settings.distributed_tracing.enabled and settings.trusted_account_key):
            return False

        if self._distributed_trace_state:
            if self._distributed_trace_state & ACCEPTED_DISTRIBUTED_TRACE:
                self._record_supportability("Supportability/DistributedTrace/AcceptPayload/Ignored/Multiple")
            else:
                self._record_supportability("Supportability/DistributedTrace/AcceptPayload/Ignored/CreateBeforeAccept")
            return False

        return True

    def _accept_distributed_trace_payload(self, payload, transport_type="HTTP"):
        if not payload:
            self._record_supportability("Supportability/DistributedTrace/AcceptPayload/Ignored/Null")
            return False

        payload = DistributedTracePayload.decode(payload)
        if not payload:
            self._record_supportability("Supportability/DistributedTrace/AcceptPayload/ParseException")
            return False

        try:
            version = payload.get("v")
            major_version = version and int(version[0])

            if major_version is None:
                self._record_supportability("Supportability/DistributedTrace/AcceptPayload/ParseException")
                return False

            if major_version > DistributedTracePayload.version[0]:
                self._record_supportability("Supportability/DistributedTrace/AcceptPayload/Ignored/MajorVersion")
                return False

            data = payload.get("d", {})
            if not all(k in data for k in DISTRIBUTED_TRACE_KEYS_REQUIRED):
                self._record_supportability("Supportability/DistributedTrace/AcceptPayload/ParseException")
                return False

            # Must have either id or tx
            if not any(k in data for k in ("id", "tx")):
                self._record_supportability("Supportability/DistributedTrace/AcceptPayload/ParseException")
                return False

            settings = self._settings
            account_id = data.get("ac")

            # If trust key doesn't exist in the payload, use account_id
            received_trust_key = data.get("tk", account_id)
            if settings.trusted_account_key != received_trust_key:
                self._record_supportability("Supportability/DistributedTrace/AcceptPayload/Ignored/UntrustedAccount")
                if settings.debug.log_untrusted_distributed_trace_keys:
                    _logger.debug(
                        "Received untrusted key in distributed trace payload. received_trust_key=%r",
                        received_trust_key,
                    )
                return False

            try:
                data["ti"] = int(data["ti"])
            except:
                return False

            if "pr" in data:
                try:
                    data["pr"] = float(data["pr"])
                except:
                    data["pr"] = None

            self._accept_distributed_trace_data(data, transport_type)
            self._record_supportability("Supportability/DistributedTrace/AcceptPayload/Success")
            return True

        except:
            self._record_supportability("Supportability/DistributedTrace/AcceptPayload/Exception")
            return False

    def accept_distributed_trace_payload(self, *args, **kwargs):
        warnings.warn(
            (
                "The accept_distributed_trace_payload API has been deprecated. "
                "Please use the accept_distributed_trace_headers API."
            ),
            DeprecationWarning,
        )
        if not self._can_accept_distributed_trace_headers():
            return False
        return self._accept_distributed_trace_payload(*args, **kwargs)

    def _accept_distributed_trace_data(self, data, transport_type):
        if transport_type not in DISTRIBUTED_TRACE_TRANSPORT_TYPES:
            transport_type = "Unknown"

        self.parent_transport_type = transport_type

        self.parent_type = data.get("ty")

        self.parent_span = data.get("id")
        self.parent_tx = data.get("tx")
        self.parent_app = data.get("ap")
        self.parent_account = data.get("ac")

        self._trace_id = data.get("tr")

        priority = data.get("pr")
        if priority is not None:
            self._priority = priority
            self._sampled = data.get("sa")

        if "ti" in data:
            transport_start = data["ti"] / 1000.0

            # If starting in the future, transport duration should be set to 0
            now = time.time()
            if transport_start > now:
                self.parent_transport_duration = 0.0
            else:
                self.parent_transport_duration = now - transport_start

        self._distributed_trace_state = ACCEPTED_DISTRIBUTED_TRACE

    def accept_distributed_trace_headers(self, headers, transport_type="HTTP"):
        if not self._can_accept_distributed_trace_headers():
            return False

        try:
            traceparent = headers.get("traceparent", "")
            tracestate = headers.get("tracestate", "")
            distributed_header = headers.get("newrelic", "")
        except Exception:
            traceparent = ""
            tracestate = ""
            distributed_header = ""

            for k, v in headers:
                k = ensure_str(k)
                if k == "traceparent":
                    traceparent = v
                elif k == "tracestate":
                    tracestate = v
                elif k == "newrelic":
                    distributed_header = v

        if traceparent:
            try:
                traceparent = ensure_str(traceparent).strip()
                data = W3CTraceParent.decode(traceparent)
            except:
                data = None

            if not data:
                self._record_supportability("Supportability/TraceContext/TraceParent/Parse/Exception")
                return False

            self._record_supportability("Supportability/TraceContext/TraceParent/Accept/Success")
            if tracestate:
                tracestate = ensure_str(tracestate)
                try:
                    vendors = W3CTraceState.decode(tracestate)
                    tk = self._settings.trusted_account_key
                    payload = vendors.pop(tk + "@nr", "")
                    self.tracing_vendors = ",".join(vendors.keys())
                    self.tracestate = vendors.text(limit=31)
                except:
                    self._record_supportability("Supportability/TraceContext/TraceState/Parse/Exception")
                else:
                    # Remove trusted new relic header if available and parse
                    if payload:
                        try:
                            tracestate_data = NrTraceState.decode(payload, tk)
                        except:
                            tracestate_data = None
                        if tracestate_data:
                            self.trusted_parent_span = tracestate_data.pop("id", None)
                            data.update(tracestate_data)
                        else:
                            self._record_supportability("Supportability/TraceContext/TraceState/InvalidNrEntry")
                    else:
                        self._record_supportability("Supportability/TraceContext/TraceState/NoNrEntry")

            self._accept_distributed_trace_data(data, transport_type)
            self._record_supportability("Supportability/TraceContext/Accept/Success")
            return True
        elif distributed_header:
            distributed_header = ensure_str(distributed_header)
            return self._accept_distributed_trace_payload(distributed_header, transport_type)

    def _process_incoming_cat_headers(self, encoded_cross_process_id, encoded_txn_header):
        settings = self._settings

        if not self.enabled:
            return

        if not (
            settings.cross_application_tracer.enabled
            and settings.cross_process_id
            and settings.trusted_account_ids
            and settings.encoding_key
        ):
            return

        if encoded_cross_process_id is None:
            return

        try:
            client_cross_process_id = deobfuscate(encoded_cross_process_id, settings.encoding_key)

            # The cross process ID consists of the client
            # account ID and the ID of the specific application
            # the client is recording requests against. We need
            # to validate that the client account ID is in the
            # list of trusted account IDs and ignore it if it
            # isn't. The trusted account IDs list has the
            # account IDs as integers, so save the client ones
            # away as integers here so easier to compare later.

            client_account_id, client_application_id = map(int, client_cross_process_id.split("#"))

            if client_account_id not in settings.trusted_account_ids:
                return

            self.client_cross_process_id = client_cross_process_id
            self.client_account_id = client_account_id
            self.client_application_id = client_application_id

            txn_header = json_decode(deobfuscate(encoded_txn_header, settings.encoding_key))

            if txn_header:
                self.is_part_of_cat = True
                self.referring_transaction_guid = txn_header[0]

                # Incoming record_tt is OR'd with existing
                # record_tt. In the scenario where we make multiple
                # ext request, this will ensure we don't set the
                # record_tt to False by a later request if it was
                # set to True by an earlier request.

                self.record_tt = self.record_tt or txn_header[1]

                if isinstance(txn_header[2], six.string_types):
                    self._trip_id = txn_header[2]
                if isinstance(txn_header[3], six.string_types):
                    self._referring_path_hash = txn_header[3]
        except Exception:
            pass

    def _generate_response_headers(self, read_length=None):
        nr_headers = []

        # Generate metrics and response headers for inbound cross
        # process web external calls.

        if self.client_cross_process_id is not None:

            # Need to work out queueing time and duration up to this
            # point for inclusion in metrics and response header. If the
            # recording of the transaction had been prematurely stopped
            # via an API call, only return time up until that call was
            # made so it will match what is reported as duration for the
            # transaction.

            if self.queue_start:
                queue_time = self.start_time - self.queue_start
            else:
                queue_time = 0

            if self.end_time:
                duration = self.end_time - self.start_time
            else:
                duration = time.time() - self.start_time

            # Generate the additional response headers which provide
            # information back to the caller. We need to freeze the
            # transaction name before adding to the header.

            self._freeze_path()

            if read_length is None:
                read_length = self._read_length

            read_length = read_length if read_length is not None else -1

            payload = (
                self._settings.cross_process_id,
                self.path,
                queue_time,
                duration,
                read_length,
                self.guid,
                self.record_tt,
            )
            app_data = json_encode(payload)

            nr_headers.append(("X-NewRelic-App-Data", obfuscate(app_data, self._settings.encoding_key)))

        return nr_headers

    def get_response_metadata(self):
        nr_headers = dict(self._generate_response_headers())
        return convert_to_cat_metadata_value(nr_headers)

    def process_request_metadata(self, cat_linking_value):
        try:
            payload = base64_decode(cat_linking_value)
        except:
            # `cat_linking_value` should always be able to be base64_decoded.
            # If this is encountered, the data being sent is corrupt. No
            # exception should be raised.
            return

        nr_headers = json_decode(payload)
        # TODO: All the external CAT APIs really need to
        # be refactored into the transaction class.
        encoded_cross_process_id = nr_headers.get("X-NewRelic-ID")
        encoded_txn_header = nr_headers.get("X-NewRelic-Transaction")
        return self._process_incoming_cat_headers(encoded_cross_process_id, encoded_txn_header)

    def set_transaction_name(self, name, group=None, priority=None):

        # Always perform this operation even if the transaction
        # is not active at the time as will be called from
        # constructor. If path has been frozen do not allow
        # name/group to be overridden. New priority then must be
        # same or greater than existing priority. If no priority
        # always override the existing name/group if not frozen.

        if self._name_priority is None:
            return

        if priority is not None and priority < self._name_priority:
            return

        if priority is not None:
            self._name_priority = priority

        # The name can be a URL for the default case. URLs are
        # supposed to be ASCII but can get a URL with illegal
        # non ASCII characters. As the rule patterns and
        # replacements are Unicode then can get Unicode
        # conversion warnings or errors when URL is converted to
        # Unicode and default encoding is ASCII. Thus need to
        # convert URL to Unicode as Latin-1 explicitly to avoid
        # problems with illegal characters.

        if isinstance(name, bytes):
            name = name.decode("Latin-1")

        # Handle incorrect groupings and leading slashes. This will
        # cause an empty segment which we want to avoid. In that case
        # insert back in Function as the leading segment.

        group = group or "Function"

        if group.startswith("/"):
            group = "Function" + group

        self._group = group
        self._name = name

    def record_exception(self, exc=None, value=None, tb=None, params=None, ignore_errors=None):
        # Deprecation Warning
        warnings.warn(
            ("The record_exception function is deprecated. Please use the new api named notice_error instead."),
            DeprecationWarning,
        )

        self.notice_error(error=(exc, value, tb), attributes=params, ignore=ignore_errors)

    def notice_error(self, error=None, attributes=None, expected=None, ignore=None, status_code=None):
        settings = self._settings

        if not settings:
            return

        if not settings.error_collector.enabled:
            return

        if not settings.collect_errors and not settings.collect_error_events:
            return

        current_span = trace_cache().current_trace()
        if current_span:
            current_span.notice_error(
                error=error,
                attributes=attributes,
                expected=expected,
                ignore=ignore,
                status_code=status_code,
            )

    def _create_error_node(self, settings, fullname, message, expected, custom_params, span_id, tb):
        # Only remember up to limit of what can be caught for a
        # single transaction. This could be trimmed further
        # later if there are already recorded errors and would
        # go over the harvest limit.

        if len(self._errors) >= settings.agent_limits.errors_per_transaction:
            return

        # Check that we have not recorded this exception
        # previously for this transaction due to multiple
        # error traces triggering. This is not going to be
        # exact but the UI hides exceptions of same type
        # anyway. Better that we under count exceptions of
        # same type and message rather than count same one
        # multiple times.

        for error in self._errors:
            if error.type == fullname and error.message == message:
                return

        node = newrelic.core.error_node.ErrorNode(
            timestamp=time.time(),
            type=fullname,
            message=message,
            expected=expected,
            span_id=span_id,
            stack_trace=exception_stack(tb),
            custom_params=custom_params,
            file_name=None,
            line_number=None,
            source=None,
        )

        # TODO Errors are recorded in time order. If
        # there are two exceptions of same type and
        # different message, the UI displays the first
        # one. In the PHP agent it was recording the
        # errors in reverse time order and so the UI
        # displayed the last one. What is the the
        # official order in which they should be sent.

        self._errors.append(node)

    def record_custom_metric(self, name, value):
        self._custom_metrics.record_custom_metric(name, value)

    def record_custom_metrics(self, metrics):
        for name, value in metrics:
            self._custom_metrics.record_custom_metric(name, value)

    def record_custom_event(self, event_type, params):
        settings = self._settings

        if not settings:
            return

        if not settings.custom_insights_events.enabled:
            return

        event = create_custom_event(event_type, params)
        if event:
            self._custom_events.add(event, priority=self.priority)

    def _intern_string(self, value):
        return self._string_cache.setdefault(value, value)

    def _process_node(self, node):
        self._trace_node_count += 1
        node.node_count = self._trace_node_count
        self.total_time += node.exclusive

        if type(node) is newrelic.core.database_node.DatabaseNode:
            settings = self._settings
            if not settings.collect_traces:
                return
            if not settings.slow_sql.enabled and not settings.transaction_tracer.explain_enabled:
                return
            if settings.transaction_tracer.record_sql == "off":
                return
            if node.duration < settings.transaction_tracer.explain_threshold:
                return
            self._slow_sql.append(node)

    def stop_recording(self):
        if not self.enabled:
            return

        if self.stopped:
            return

        if self.end_time:
            return

        self.end_time = time.time()
        self.stopped = True

        if self._utilization_tracker:
            if self._thread_utilization_start:
                if not self._thread_utilization_end:
                    self._thread_utilization_end = self._utilization_tracker.utilization_count()

        self._cpu_user_time_end = os.times()[0]

    def add_custom_parameter(self, name, value):
        if not self._settings:
            return False

        if self._settings.high_security:
            _logger.debug("Cannot add custom parameter in High Security Mode.")
            return False

        if len(self._custom_params) >= MAX_NUM_USER_ATTRIBUTES:
            _logger.debug("Maximum number of custom attributes already added. Dropping attribute: %r=%r", name, value)
            return False

        key, val = process_user_attribute(name, value)

        if key is None:
            return False
        else:
            self._custom_params[key] = val
            return True

    def add_custom_parameters(self, items):
        result = True

        # items is a list of (name, value) tuples.
        for name, value in items:
            result &= self.add_custom_parameter(name, value)

        return result

    def add_framework_info(self, name, version=None):
        if name:
            self._frameworks.add((name, version))

    def dump(self, file):
        """Dumps details about the transaction to the file object."""

        print("Application: %s" % (self.application.name), file=file)
        print("Time Started: %s" % (time.asctime(time.localtime(self.start_time))), file=file)
        print("Thread Id: %r" % (self.thread_id), file=file)
        print("Current Status: %d" % (self._state), file=file)
        print("Recording Enabled: %s" % (self.enabled), file=file)
        print("Ignore Transaction: %s" % (self.ignore_transaction), file=file)
        print("Transaction Dead: %s" % (self._dead), file=file)
        print("Transaction Stopped: %s" % (self.stopped), file=file)
        print("Background Task: %s" % (self.background_task), file=file)
        print("Request URI: %s" % (self._request_uri), file=file)
        print("Transaction Group: %s" % (self._group), file=file)
        print("Transaction Name: %s" % (self._name), file=file)
        print("Name Priority: %r" % (self._name_priority), file=file)
        print("Frozen Path: %s" % (self._frozen_path), file=file)
        print("AutoRUM Disabled: %s" % (self.autorum_disabled), file=file)
        print("Supress Apdex: %s" % (self.suppress_apdex), file=file)


def current_transaction(active_only=True):
    current = trace_cache().current_transaction()
    if active_only:
        if current and (current.ignore_transaction or current.stopped):
            return None
    return current


def set_transaction_name(name, group=None, priority=None):
    transaction = current_transaction()
    if transaction:
        transaction.set_transaction_name(name, group, priority)


def end_of_transaction():
    transaction = current_transaction()
    if transaction:
        transaction.stop_recording()


def set_background_task(flag=True):
    transaction = current_transaction()
    if transaction:
        transaction.background_task = flag


def ignore_transaction(flag=True):
    transaction = current_transaction()
    if transaction:
        transaction.ignore_transaction = flag


def suppress_apdex_metric(flag=True):
    transaction = current_transaction()
    if transaction:
        transaction.suppress_apdex = flag


def capture_request_params(flag=True):
    transaction = current_transaction()
    if transaction and transaction.settings:
        if transaction.settings.high_security:
            _logger.warn("Cannot modify capture_params in High Security Mode.")
        else:
            transaction.capture_params = flag


def add_custom_parameter(key, value):
    transaction = current_transaction()
    if transaction:
        return transaction.add_custom_parameter(key, value)
    else:
        return False


def add_custom_parameters(items):
    transaction = current_transaction()
    if transaction:
        return transaction.add_custom_parameters(items)
    else:
        return False


def add_framework_info(name, version=None):
    transaction = current_transaction()
    if transaction:
        transaction.add_framework_info(name, version)


def get_browser_timing_header():
    transaction = current_transaction()
    if transaction and hasattr(transaction, "browser_timing_header"):
        return transaction.browser_timing_header()
    return ""


def get_browser_timing_footer():
    transaction = current_transaction()
    if transaction and hasattr(transaction, "browser_timing_footer"):
        return transaction.browser_timing_footer()
    return ""


def disable_browser_autorum(flag=True):
    transaction = current_transaction()
    if transaction:
        transaction.autorum_disabled = flag


def suppress_transaction_trace(flag=True):
    transaction = current_transaction()
    if transaction:
        transaction.suppress_transaction_trace = flag


def record_custom_metric(name, value, application=None):
    if application is None:
        transaction = current_transaction()
        if transaction:
            transaction.record_custom_metric(name, value)
        else:
            _logger.debug(
                "record_custom_metric has been called but no "
                "transaction was running. As a result, the following metric "
                "has not been recorded. Name: %r Value: %r. To correct this "
                "problem, supply an application object as a parameter to this "
                "record_custom_metrics call.",
                name,
                value,
            )
    elif application.enabled:
        application.record_custom_metric(name, value)


def record_custom_metrics(metrics, application=None):
    if application is None:
        transaction = current_transaction()
        if transaction:
            transaction.record_custom_metrics(metrics)
        else:
            _logger.debug(
                "record_custom_metrics has been called but no "
                "transaction was running. As a result, the following metrics "
                "have not been recorded: %r. To correct this problem, "
                "supply an application object as a parameter to this "
                "record_custom_metric call.",
                list(metrics),
            )
    elif application.enabled:
        application.record_custom_metrics(metrics)


def record_custom_event(event_type, params, application=None):
    """Record a custom event.

    Args:
        event_type (str): The type (name) of the custom event.
        params (dict): Attributes to add to the event.
        application (newrelic.api.Application): Application instance.

    """

    if application is None:
        transaction = current_transaction()
        if transaction:
            transaction.record_custom_event(event_type, params)
        else:
            _logger.debug(
                "record_custom_event has been called but no "
                "transaction was running. As a result, the following event "
                "has not been recorded. event_type: %r params: %r. To correct "
                "this problem, supply an application object as a parameter to "
                "this record_custom_event call.",
                event_type,
                params,
            )
    elif application.enabled:
        application.record_custom_event(event_type, params)


def accept_distributed_trace_payload(payload, transport_type="HTTP"):
    transaction = current_transaction()
    if transaction:
        return transaction.accept_distributed_trace_payload(payload, transport_type)
    return False


def accept_distributed_trace_headers(headers, transport_type="HTTP"):
    transaction = current_transaction()
    if transaction:
        return transaction.accept_distributed_trace_headers(headers, transport_type)


def create_distributed_trace_payload():
    transaction = current_transaction()
    if transaction:
        return transaction.create_distributed_trace_payload()


def insert_distributed_trace_headers(headers):
    transaction = current_transaction()
    if transaction:
        return transaction.insert_distributed_trace_headers(headers)


def current_trace_id():
    transaction = current_transaction()
    if transaction:
        return transaction.trace_id


def current_span_id():
    trace = trace_cache().current_trace()
    if trace:
        return trace.guid



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from logging import Formatter, LogRecord

from newrelic.api.time_trace import get_linking_metadata
from newrelic.common.object_names import parse_exc_info
from newrelic.core.config import is_expected_error


def format_exc_info(exc_info):
    _, _, fullnames, message = parse_exc_info(exc_info)
    fullname = fullnames[0]

    formatted = {
        "error.class": fullname,
        "error.message": message,
    }

    expected = is_expected_error(exc_info)
    if expected is not None:
        formatted["error.expected"] = expected

    return formatted


class NewRelicContextFormatter(Formatter):
    DEFAULT_LOG_RECORD_KEYS = frozenset(vars(LogRecord("", 0, "", 0, "", (), None)))

    def __init__(self, *args, **kwargs):
        super(NewRelicContextFormatter, self).__init__()

    @classmethod
    def log_record_to_dict(cls, record):
        output = {
            "timestamp": int(record.created * 1000),
            "message": record.getMessage(),
            "log.level": record.levelname,
            "logger.name": record.name,
            "thread.id": record.thread,
            "thread.name": record.threadName,
            "process.id": record.process,
            "process.name": record.processName,
            "file.name": record.pathname,
            "line.number": record.lineno,
        }
        output.update(get_linking_metadata())

        DEFAULT_LOG_RECORD_KEYS = cls.DEFAULT_LOG_RECORD_KEYS
        if len(record.__dict__) > len(DEFAULT_LOG_RECORD_KEYS):
            for key in record.__dict__:
                if key not in DEFAULT_LOG_RECORD_KEYS:
                    output["extra." + key] = getattr(record, key)

        if record.exc_info:
            output.update(format_exc_info(record.exc_info))

        return output

    def format(self, record):
        def safe_str(object, *args, **kwargs):
            """Convert object to str, catching any errors raised."""
            try:
                return str(object, *args, **kwargs)
            except:
                return "<unprintable %s object>" % type(object).__name__

        return json.dumps(self.log_record_to_dict(record), default=safe_str, separators=(",", ":"))
